#  Copyright (c) 2024 zfit
import pytest
import tensorflow as tf
import zfit
from zfit import z

# this works now also for tensorflow, as all variables given as args, effectively Params, have a
# custom tf retracing behavior on their identity
# Therefore, the legacy fix in z.function works (still, but z.function is not needed anymore for this)
# and with tf.function
@pytest.mark.parametrize("function", [z.function, tf.function])
def test_parameter_caching(function):
    # this is to ensure that we fixed the bug (https://github.com/tensorflow/tensorflow/issues/57365) internally

    x1 = zfit.Parameter("x1", 2.0)
    x2 = zfit.Parameter("x2", 4.0)


    def one_plus(x):
        return x + 1

    cx1 = zfit.ComposedParameter("cx1", one_plus, dependents=[x1])
    cx2 = zfit.ComposedParameter("cx2", one_plus, dependents=[x2])

    ncompile1 = 0
    ncompile2 = 0


    @tf.function(autograph=False)
    def f():
        nonlocal ncompile1
        ncompile1 += 1
        res = x1 + x2**2 / 2
        return res

    def f2():
        nonlocal ncompile2
        ncompile2 += 1
        res = cx1 + cx2**2 / 2
        return res

    ncompgrad1 = 0
    ncompgrad2 = 0
    def grad(param):
        nonlocal ncompgrad1
        ncompgrad1 += 1

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(param)
            value = f()
        return tf.stack(tape.gradient(value, param))

    def grad2(param):
        nonlocal ncompgrad2
        ncompgrad2 += 1
        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(param)
            value = f2()
        return tf.stack(tape.gradient(value, param))

    jitted_grad = function(grad)
    jitted_grad2 = function(grad2)

    y1 = grad(x1)
    ncomp1_after1 = ncompile1
    y1_jit = jitted_grad(x1)
    ncompgrad1_after1 = ncompgrad1
    assert ncomp1_after1 == ncompile1
    y1_jit = jitted_grad(x1)
    assert ncompgrad1_after1 == ncompgrad1
    with x1.set_value(142.):
        y1_jit = jitted_grad(x1)
        assert ncompgrad1_after1 == ncompgrad1

    y1_jit = jitted_grad(x1)
    assert ncompgrad1_after1 == ncompgrad1

    assert abs(y1 - 1.0) < 1e-5  # because d x1 / dx1 = 1
    assert abs(y1_jit - 1.0) < 1e-5
    y2 = grad(x2)
    y2_jit = jitted_grad(x2)

    assert abs(y2 - 4.0) < 1e-5  # because d / dx x**2/2 = x -> 4
    assert abs(y2_jit - y2) < 1e-5
    assert abs(y2_jit - 4.0) < 1e-5

    # use both parameters
    y = grad([x1, x2])
    ncompgrad1_after2 = ncompgrad1
    assert ncompgrad1_after2 > ncompgrad1_after1
    y_jit = jitted_grad([x1, x2])

    assert abs(y[0] - 1.0) < 1e-5
    assert abs(y[1] - 4.0) < 1e-5
    assert abs(y_jit[0] - 1.0) < 1e-5
    assert abs(y_jit[1] - 4.0) < 1e-5

    # use both parameters, swap order
    y = grad([x2, x1])

    y_jit = jitted_grad([x2, x1])

    assert pytest.approx(y[0], 1e-5) == 4.0
    assert pytest.approx(y[1], 1e-5) == 1.0
    assert pytest.approx(y_jit[0], 1e-5) == 4.0
    assert pytest.approx(y_jit[1], 1e-5) == 1.0

    # test with composed parameters
    y = grad2([x1, x2])
    y_jit = jitted_grad2([x1, x2])

    assert pytest.approx(y[0], 1e-5) == y_jit[0]
    assert pytest.approx(y[1], 1e-5) == y_jit[1]

    # test with composed parameters, swap order
    y = grad2([x2, x1])
    y_jit = jitted_grad2([x2, x1])

    assert pytest.approx(y[0], 1e-5) == y_jit[0]
    assert pytest.approx(y[1], 1e-5) == y_jit[1]



#  Copyright (c) 2024 zfit
from __future__ import annotations

import zfit

# create space
obs = zfit.Space("x", -10, 10)

# parameters
mu = zfit.Parameter("mu", 1.0, -4, 6)
sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
frac = zfit.Parameter("fraction", 0.3, 0, 1)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
exponential = zfit.pdf.Exponential(lambd, obs=obs)
model = zfit.pdf.SumPDF([gauss, exponential], fracs=frac)

# data
n_sample = 10000

# if we sample once, otherwise use create_sampler
data = model.sample(n_sample, limits=obs)  # limits can be omitted, then the default limits are used

# set the values to a start value for the fit
zfit.param.set_values({mu: 0.5, sigma: 1.2, lambd: -0.05, frac: 0.07})

# create NLL
nll = zfit.loss.UnbinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit()
zfit.run.experimental_disable_param_update()
result = minimizer.minimize(
    nll
).update_params()  # forward compatibility, update_params will update the values of the parameters to the minimum

# do the error calculations, here with minos
param_hesse = result.hesse()
param_errors, new_result = result.errors()



#  Copyright (c) 2024 zfit

from __future__ import annotations

from .util.plotter import PDFPlotter, SumCompPlotter, ZfitPDFPlotter, plot_model_pdfV1, plot_sumpdf_components_pdfV1

__all__ = ["plot_model_pdfV1", "plot_sumpdf_components_pdfV1", "ZfitPDFPlotter", "PDFPlotter", "SumCompPlotter"]



#  Copyright (c) 2024 zfit
from __future__ import annotations

from typing import Callable, Mapping, Optional

from ..core.interfaces import ZfitPDF
from ..core.space import convert_to_space
from . import ztyping
from .checks import RuntimeDependency
from .exception import WorkInProgressError
from .warnings import warn_experimental_feature

try:
    import matplotlib.pyplot as plt
except ImportError as error:
    plt = RuntimeDependency("plt", error_msg=str(error))


def plot_sumpdf_components_pdfV1(
    model,
    *,
    plotfunc: Optional[Callable] = None,
    scale=1,
    ax=None,
    linestyle=None,
    plotkwargs: Mapping[str, object] | None = None,
):
    """Plot the components of a sum pdf.

    Args:
        model: A zfit SumPDF.
        plotfunc: A plotting function that takes the `ax` to plot on x, y, and additional arguments.
        scale: An overall scale factor to apply to the components.
        ax: A matplotlib Axes object to plot on.
        linestyle: A linestyle to use for the components. Default is "--".
        plotkwargs: Additional keyword arguments to pass to the plotting function.
    """
    import zfit

    if not isinstance(model, zfit.pdf.SumPDF):
        msg = f"model must be a ZfitPDF, not a {type(model)}. Model is {model}."
        raise ValueError(msg)
    if linestyle is None:
        linestyle = "--"
    if plotkwargs is None:
        plotkwargs = {}
    plotfunc = plot_model_pdfV1 if plotfunc is None else plotfunc
    for mod, frac in zip(model.pdfs, model.params.values()):
        plotfunc(mod, scale=frac * scale, ax=ax, linestyle=linestyle, plotkwargs=plotkwargs)
    return ax


def plot_model_pdfV1(
    model: ZfitPDF,
    *,
    plotfunc: Callable | None = None,
    extended: bool | None = None,
    obs: ztyping.ObsTypeInput = None,
    scale: float | int | None = None,
    ax: plt.Axes | None = None,
    num: int | None = None,
    full: bool | None = None,
    linestyle=None,
    plotkwargs=None,
):
    """Plot the 1 dimensional density of a model, possibly scaled by the yield if extended.

    Args:
        model: An unbinned ZfitPDF.
        plotfunc: A plotting function that takes the ``ax`` to plot on, and  x, y, and additional arguments. Default is ``ax.plot``.
        extended: If True, plot the extended pdf. If False, plot the pdf.
        obs: The observable to plot the pdf for. If None, the model's space is used.
        scale: An overall scale factor to apply to the pdf.
        ax: A matplotlib Axes object to plot on.
        num: The number of points to evaluate the pdf at. Default is 300.
        full: If True, set the x and y labels and the legend. Default is True.
        linestyle: A linestyle to use for the pdf.
        plotkwargs: Additional keyword arguments to pass to the plotting function.

    Returns:
    """
    import zfit.z.numpy as znp

    if not isinstance(model, ZfitPDF):
        msg = f"model must be a ZfitPDF, not a {type(model)}. Model is {model}."
        raise TypeError(msg)

    if extended is None:
        extended = model.is_extended

    if scale is None:
        scale = 1
    if num is None:
        num = 300
    if full is None:
        full = True
    if plotkwargs is None:
        plotkwargs = {}

    if obs is None:
        obs = model.space
        if not obs.has_limits:
            msg = "Observables must have limits to be plotted. Either provide the limits with `obs` or use a model that has limits."
            raise ValueError(msg)
    else:
        obs = convert_to_space(obs)
        if not obs.has_limits:
            obs = model.space.with_obs(obs)
        if not obs.has_limits:
            msg = "Observables must have limits to be plotted. Either provide the limits with `obs` or use a model that has limits."
            raise ValueError(msg)

    if obs.n_obs != 1:
        msg = "obs must be 1D to be plotted."
        raise ValueError(msg)
    if model.space.n_obs != 1:
        if obs is None:
            msg = "1D space must be provided for multi-dimensional models to provide a 1D projection."
            raise ValueError(msg)
        model = model.create_projection_pdf(obs=obs, label=model.label)
    lower, upper = obs.v1.limits
    x = znp.linspace(lower, upper, num=num)
    y = model.ext_pdf(x) if extended else model.pdf(x)
    y *= scale
    if ax is None:
        ax = plt.gca()
    elif not isinstance(ax, plt.Axes):
        msg = "ax must be a matplotlib Axes object"
        raise ValueError(msg)
    plotfunc = ax.plot if plotfunc is None else plotfunc
    if "label" not in plotkwargs and full:
        plotkwargs["label"] = model.label
    plotfunc(x, y, linestyle=linestyle, **plotkwargs)
    if full:
        ax.set_xlabel(obs.label)
        ylabel = "Probability density" if not extended else "Extended probability density"
        ax.set_ylabel(ylabel)
        plt.legend()
    return ax


def assert_initialized(func):
    def wrapper(self, *args, **kwargs):
        if self.pdf is None:
            msg = "PDFPlotter is not initialized with a PDF."
            raise ValueError(msg)
        return func(self, *args, **kwargs)

    return wrapper


class ZfitPDFPlotter:
    @warn_experimental_feature
    @assert_initialized
    def plotpdf(
        self,
        *,
        plotfunc: Callable | None = None,
        extended: bool | None = None,
        obs: ztyping.ObsTypeInput = None,
        scale: float | int | None = None,
        ax: plt.Axes | None = None,
        num: int | None = None,
        full: bool | None = None,
        linestyle=None,
        plotkwargs=None,
    ):
        """Plot the 1 dimensional density of the PDF, possibly scaled by the yield if extended.

        Args:
            plotfunc: A plotting function that takes the `ax` to plot on, and x, y, and additional arguments. Default is `ax.plot`.
            extended: If True, plot the extended pdf. If False, plot the pdf.
            obs: The observable to plot the pdf for. If None, the model's space is used.
            scale: An overall scale factor to apply to the pdf.
            ax: A matplotlib Axes object to plot on.
            num: The number of points to evaluate the pdf at. Default is 300.
            full: If True, set the x and y labels and the legend. Default is True.
            linestyle: A linestyle to use for the pdf.
            plotkwargs: Additional keyword arguments to pass to the plotting function.

        Returns:
            The matplotlib Axes object.
        """
        return self._plotpdf(
            plotfunc=plotfunc,
            extended=extended,
            obs=obs,
            scale=scale,
            ax=ax,
            num=num,
            full=full,
            linestyle=linestyle,
            plotkwargs=plotkwargs,
        )

    def _plotpdf(self, **kwargs):
        raise NotImplementedError

    @property
    def comp(self):
        return None


class PDFPlotter(ZfitPDFPlotter):
    def __init__(
        self,
        pdf: ZfitPDF | None,
        pdfplotter: Callable | None = None,
        componentplotter: ZfitPDFPlotter = None,
        defaults: Mapping[str, object] | None = None,
    ):
        self.defaults = {} if defaults is None else defaults
        self.pdf = pdf
        if pdfplotter is not None and not callable(pdfplotter):
            msg = f"pdfplotter must be a callable, is {type(pdfplotter)}."
            raise TypeError(msg)
        self._pdfplotter = plot_model_pdfV1 if pdfplotter is None else pdfplotter
        if componentplotter is not None and not isinstance(componentplotter, ZfitPDFPlotter):
            msg = f"componentplotter must be a ZfitPDFPlotter, is {type(componentplotter)}."
            raise TypeError(msg)
        self._componentplotter = componentplotter

    def __call__(self):
        msg = "This is not yet implemented. Use the methods instead."
        raise WorkInProgressError(msg)

    def _plotpdf(self, **kwargs):
        kwargs |= self.defaults
        return plot_model_pdfV1(self.pdf, **kwargs)

    @property
    @assert_initialized
    def comp(self):
        return self._componentplotter


class SumCompPlotter(ZfitPDFPlotter):
    def __init__(
        self,
        pdf: ZfitPDF | None,
        *args,
        **kwargs,
    ):
        if pdf is not None and not isinstance(pdf, ZfitPDF):
            msg = f"pdf must be a ZfitPDF, is {type(pdf)}."
            raise TypeError(msg)
        self.pdf = pdf
        super().__init__(*args, **kwargs)

    def _plotpdf(self, **kwargs):
        import zfit

        if not isinstance(pdf := self.pdf, zfit.pdf.SumPDF):  # we can relax this later with duck typing
            msg = f"pdf must be a SumPDF, is {type(pdf)}."
            raise TypeError(msg)
        scale = kwargs.pop("scale")
        if scale is None:
            scale = 1
        for mod, frac in zip(pdf.pdfs, pdf.params.values()):
            ax = mod.plot.plotpdf(scale=frac * scale, **kwargs)
        return ax



"""This  module defines the ``BasePdf`` that can be used to inherit from in order to build a custom PDF.

The ``BasePDF`` implements already a lot of ready-to-use functionality like integral, automatic normalization
and sampling.

Defining your own pdf
---------------------

A simple example:
>>> import zfit
>>> import zfit.z.numpy as znp
>>>
>>> class MyGauss(BasePDF):
>>>     def __init__(self, mean, stddev, name="MyGauss"):
>>>         super().__init__(mean=mean, stddev=stddev, name=name)
>>>
>>>     def _unnormalized_pdf(self, x):
>>>         return znp.exp((x - mean) ** 2 / (2 * stddev**2))

Notice that *here* we only specify the *function* and no normalization. This
**No** attempt to **explicitly** normalize the function should be done inside ``_unnormalized_pdf``.
The normalization is handled with another method depending on the normalization range specified.
(It *is* possible, though discouraged, to directly provide the *normalized probability* by overriding _pdf(), but
there are other, more convenient ways to add improvements like providing an analytical integrals.)

Before we create an instance, we need to create the variables to initialize it
>>> mean = zfit.Parameter("mean1", 2., 0.1, 4.2)  # signature as in RooFit: *name, initial, lower, upper*
>>> stddev = zfit.Parameter("stddev1", 5., 0.3, 10.)
Let's create an instance and some example data
>>> gauss = MyGauss(mean=mean, stddev=stddev)
>>> example_data = np.random.random(10)
Now we can get the probability
>>> probs = gauss.pdf(example_data)  # ``norm`` specifies over which range to normalize
Or the integral
>>> integral = gauss.integrate(limits=(-5, 3.1),norm=False)  # norm_range is False -> return unnormalized
integral
Or directly sample from it
>>> sample = gauss.sample(n_draws=1000, limits=(-10, 10))  # draw 1000 samples within (-10, 10)

We can create an extended PDF, which will result in anything using a ``norm_range`` to not return the
probability but the number probability (the function will be normalized to ``yield`` instead of 1 inside
the ``norm_range``)
>>> yield1 = Parameter("yield1", 100, 0, 1000)
>>> gauss_extended = gauss.create_extended(yield1)
>>> gauss.is_extended
True

>>> integral_extended = gauss.ext_integrate(limits=(-10, 10),norm=(-10, 10))  # yields approx 100

For more advanced methods and ways to register analytic integrals or overwrite certain methods, see
also the advanced models in `zfit models <https://github.com/zfit/zfit-tutorials>`_
"""

#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional

from tensorflow.python.util.deprecation import deprecated_args

from ..util.plotter import PDFPlotter
from ..util.ztyping import ExtendedInputType, NormInputType

if TYPE_CHECKING:
    pass

import warnings
from contextlib import suppress

import tensorflow as tf

import zfit.z.numpy as znp
from zfit import z

from ..settings import run, ztypes
from ..util import ztyping
from ..util.cache import invalidate_graph
from ..util.deprecation import deprecated, deprecated_norm_range
from ..util.exception import (
    AlreadyExtendedPDFError,
    BreakingAPIChangeError,
    FunctionNotImplemented,
    NormNotImplemented,
    NotExtendedPDFError,
    SpecificFunctionNotImplemented,
)
from ..util.temporary import TemporarilySet
from .basemodel import BaseModel
from .baseobject import extract_filter_params
from .interfaces import ZfitParameter, ZfitPDF, ZfitSpace
from .parameter import Parameter, convert_to_parameter
from .sample import extended_sampling
from .space import Space, convert_to_space

_BasePDF_USER_IMPL_METHODS_TO_CHECK = {}


def _BasePDF_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the ``@supports`` decorator.

    Args:
        has_support: If True, flags that it **requires** the ``@supports`` decorator. If False,
            flags that the ``@supports`` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        msg = "Has to be boolean."
        raise TypeError(msg)

    def register(func):
        """Register a method to be checked to (if True) *has* ``support`` or (if False) has *no* ``support``.

        Args:
            func:

        Returns:
            Function:
        """
        name = func.__name__
        _BasePDF_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BasePDF_register_check_support
        return func

    return register


class PDFMeta(type):
    def __call__(cls, *args, obs=None, **kwargs):
        if binned := (obs is not None and isinstance(obs, Space) and obs.binning is not None):
            binned_obs = obs
            obs = binned_obs.with_binning(None)
        if obs is not None:
            kwargs["obs"] = obs
        pdf = cls.__new__(cls)
        pdf.__init__(*args, **kwargs)
        if binned:
            pdf = pdf.to_binned(
                binned_obs,
                extended=kwargs.get("extended", None),
                norm=kwargs.get("norm", None),
                name=kwargs.get("name", None),
                label=kwargs.get("label", None),
            )

        return pdf


class BasePDF(ZfitPDF, BaseModel, metaclass=PDFMeta):
    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        params: dict[str, ZfitParameter] | None = None,
        *,
        dtype=ztypes.float,
        label=None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "BasePDF",
        **kwargs,
    ):
        self._yield = None
        self.plot = None

        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)
        self._label = label or self.name
        self._norm = norm
        if extended is not False and extended is not None:
            self._set_yield(extended)

        self._assert_params_unique()
        if self.plot is None:
            self.plot = PDFPlotter(self)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._subclass_check_support(
            methods_to_check=_BasePDF_USER_IMPL_METHODS_TO_CHECK,
            wrapper_not_overwritten=_BasePDF_register_check_support,
        )

    def _check_input_norm(self, norm, none_is_error=False):
        if norm is None:
            norm = self.norm
        return super()._check_input_norm(norm=norm, none_is_error=none_is_error)

    def _check_input_params_tfp(self, *params):
        return tuple(convert_to_parameter(p) for p in params)

    def _func_to_integrate(self, x: ztyping.XType):
        return self.pdf(x, norm=False)

    def _func_to_sample_from(self, x):
        return self.pdf(x, norm=False)

    @property
    def label(self):
        return self._label

    @property
    @deprecated(None, "Use the `norm` attribute instead.")
    def norm_range(self) -> Space | None | bool:
        """Return the current normalization range. If None and the ``obs`` have limits, they are returned.

        Returns:
            The current normalization range.
        """
        return self.norm

    @property
    def norm(self) -> Space | None | bool:
        """Return the current normalization range. If None and the ``obs`` have limits, they are returned.

        Returns:
            The current normalization range.
        """
        norm = self._norm
        if norm is None:
            norm = self.space
        return norm

    @invalidate_graph
    @deprecated(None, "Prefer to create a new PDF with `norm` set or wrap the existing in a `TruncatedPDF`.")
    def set_norm_range(self, norm: ztyping.LimitsTypeInput):
        """Set the normalization range (temporarily if used with contextmanager).

        Args:
            norm:
        """
        norm = self._check_input_norm(norm)

        def setter(value):
            self._norm = value

        def getter():
            return self._norm

        return TemporarilySet(value=norm, setter=setter, getter=getter)

    @_BasePDF_register_check_support(True)
    def _normalization(self, norm, options, *, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @deprecated_args(None, "Use `norm` instead.", "limits")
    def normalization(
        self,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
        limits: ztyping.LimitsType = None,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Return the normalization of the function (usually the integral over ``norm``).

        Args:
            norm:  |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
            options: |@doc:pdf.param.options||@docend:pdf.param.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            The normalization value
        """
        del limits
        if options is None:
            options = {}
        norm = self._check_input_norm(norm)
        with self._check_set_input_params(params=params):
            return self._single_hook_normalization(norm=norm, options=options)

    def _single_hook_normalization(self, norm, options):  # TODO(Mayou36): add yield?
        return self._hook_normalization(norm=norm, options=options)

    def _hook_normalization(self, norm, options):
        return self._call_normalization(norm=norm, options=options)  # no _norm_* needed

    def _call_normalization(self, norm, options):
        # TODO: caching? alternative
        with suppress(FunctionNotImplemented):
            return self._normalization(norm=norm, options=options)
        return self._fallback_normalization(norm, options=options)

    def _fallback_normalization(self, norm, options):
        return self._hook_integrate(limits=norm, norm=False, options=options)

    def _unnormalized_pdf(self, x, *, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @deprecated(None, "Use `pdf(norm=False)` instead")
    def unnormalized_pdf(self, x: ztyping.XType) -> ztyping.XType:
        """PDF "unnormalized". Use ``functions`` for unnormalized pdfs. this is only for performance in special cases.

        Args:
            x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
        Returns:
            1-dimensional :py:class:`tf.Tensor` containing the unnormalized pdf.
        """
        with self._convert_sort_x(x) as x:
            return self._single_hook_unnormalized_pdf(x)

    def _single_hook_unnormalized_pdf(self, x):
        return self._call_unnormalized_pdf(x=x)

    def _call_unnormalized_pdf(self, x):
        # try:

        return self._unnormalized_pdf(x)

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Probability density function scaled by yield, normalized over ``norm_range``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
          params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        norm = self._check_input_norm(norm, none_is_error=True)
        if not self.is_extended:
            msg = f"{self} is not extended, cannot call `ext_pdf`"
            raise NotExtendedPDFError(msg)
        with self._convert_sort_x(x) as x, self._check_set_input_params(params=params):
            return self._call_ext_pdf(x, norm)

    def _call_ext_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_pdf(x, norm)

        # fallback
        return self.pdf(x=x, norm=norm) * self.get_yield()

    def _auto_ext_pdf(self, x, norm):
        try:
            probs = self._ext_pdf(x, norm)
        except NormNotImplemented:
            unnorm_probs = self._ext_pdf(x, False)
            normalization = self.normalization(norm)
            probs = unnorm_probs / normalization
        return probs

    @_BasePDF_register_check_support(True)
    def _ext_pdf(self, x, norm, *, norm_range=None, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented  # TODO: implement properly

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_log_pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Log of probability density function scaled by yield, normalized over ``norm_range``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|

          params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        norm = self._check_input_norm(norm, none_is_error=True)
        if not self.is_extended:
            msg = f"{self} is not extended, cannot call `ext_pdf`"
            raise NotExtendedPDFError(msg)
        with self._convert_sort_x(x) as x, self._check_set_input_params(params=params):
            return self._call_ext_log_pdf(x, norm)

    def _call_ext_log_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_log_pdf(x, norm)

        # fallback
        return self.log_pdf(x=x, norm=norm) + znp.log(self.get_yield())

    def _auto_ext_log_pdf(self, x, norm):
        try:
            pdf = self._ext_log_pdf(x, norm)
        except NormNotImplemented:
            unnormed_pdf = self._ext_log_pdf(x, False)
            normalization = self.log_normalization(norm)
            pdf = unnormed_pdf - normalization
        return pdf

    @_BasePDF_register_check_support(True)
    def _ext_log_pdf(self, x, norm):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @_BasePDF_register_check_support(True)
    def _pdf(self, x, norm, *, norm_range=None, params=None):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @z.function(wraps="model")
    @deprecated_norm_range
    def pdf(
        self,
        x: ztyping.XTypeInput,
        norm: ztyping.LimitsTypeInput = None,
        *,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Probability density function of ``x``, normalized over ``norm``.

        Args:
          x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
          norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
          params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|
        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        norm = self._check_input_norm(norm, none_is_error=True)
        with self._convert_sort_x(x) as x, self._check_set_input_params(params=params):
            value = self._single_hook_pdf(x=x, norm=norm)
        if run.numeric_checks:
            z.check_numerics(value, message="Check if pdf output contains any NaNs of Infs")
        return znp.atleast_1d(znp.asarray(z.to_real(value)))

    def _single_hook_pdf(self, x, norm):
        return self._hook_pdf(x=x, norm=norm)

    def _hook_pdf(self, x, norm):
        return self._norm_pdf(x=x, norm=norm)

    def _norm_pdf(self, x, norm):
        try:
            return self._call_pdf(x=x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._call_pdf(x=x, norm=False)
            normalization = self.normalization(norm)
            return unnormed_pdf / normalization

    def _call_pdf(self, x, norm):
        with suppress(FunctionNotImplemented):
            return self._pdf(x, norm)
        with suppress(FunctionNotImplemented):
            return znp.exp(self._log_pdf(x, norm))
        if self.is_extended:
            with suppress(FunctionNotImplemented):
                return self._ext_pdf(x, norm) / self.get_yield()  # TODO: extend/refactor the calling

        return self._fallback_pdf(x, norm)

    def _fallback_pdf(self, x, norm):
        pdf = self._call_unnormalized_pdf(x)
        if norm.has_limits:
            pdf /= self._hook_normalization(norm=norm, options={})
        return pdf

    @_BasePDF_register_check_support(True)
    @deprecated_norm_range
    def _log_pdf(self, x, norm):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    @deprecated_norm_range
    def log_pdf(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        *,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Log probability density function normalized over ``norm_range``.

        Args:
            x: |@doc:pdf.param.x| Data to evaluate the method on. Should be ``ZfitData``
               or a mapping of *obs* to numpy-like arrays.
               If an array is given, the first dimension is interpreted as the events while
               the second is meant to be the dimensionality of a single event. |@docend:pdf.param.x|
            norm: |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|
        Returns:
          A ``Tensor`` of type ``self.dtype``.
        """
        norm = self._check_input_norm(norm)
        with self._convert_sort_x(x) as x, self._check_set_input_params(params=params):
            return znp.asarray(z.to_real(self._single_hook_log_pdf(x=x, norm=norm)))

    def _single_hook_log_pdf(self, x, norm):
        return self._hook_log_pdf(x=x, norm=norm)

    def _hook_log_pdf(self, x, norm):
        return self._norm_log_pdf(x=x, norm=norm)

    def _norm_log_pdf(self, x, norm):
        try:
            return self._call_log_pdf(x=x, norm=norm)
        except NormNotImplemented:
            unnormed_log_pdf = self._call_log_pdf(x=x, norm=False)
            normalization = self.log_normalization(norm)
            return unnormed_log_pdf - normalization

    def _call_log_pdf(self, x, norm):
        with suppress(FunctionNotImplemented):
            return self._log_pdf(x, norm)
        with suppress(FunctionNotImplemented):
            return znp.log(self._pdf(x, norm))
        return self._fallback_log_pdf(x, norm)

    def _fallback_log_pdf(self, x, norm):
        return znp.log(self._hook_pdf(x=x, norm=norm))

    @_BasePDF_register_check_support(True)
    @deprecated_norm_range
    def _log_normalization(self, norm, *, params=None, options):  # noqa: ARG002
        raise SpecificFunctionNotImplemented

    def log_normalization(
        self,
        norm: ztyping.LimitsType,
        *,
        options=None,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Return the normalization of the function (usually the integral over ``norm``).

        Args:
            norm:  |@doc:pdf.param.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). Should be ``ZfitSpace`` to define the space
               to normalize over. |@docend:pdf.param.norm|
            options: |@doc:pdf.param.options||@docend:pdf.param.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            The normalization value
        """
        if options is None:
            options = {}
        norm = self._check_input_norm(norm, none_is_error=True)
        with self._check_set_input_params(params=params):
            return self._single_hook_log_normalization(norm=norm, options=options)

    def _single_hook_log_normalization(self, norm, options):  # TODO(Mayou36): add yield?
        return self._hook_normalization(norm=norm, options=options)

    def _hook_log_normalization(self, norm, options):
        return self._call_normalization(norm=norm, options=options)  # no _norm_* needed

    def _call_log_normalization(self, norm, options):
        # TODO: caching? alternative
        with suppress(FunctionNotImplemented):
            return self._normalization(norm=norm, options=options)
        return self._fallback_normalization(norm, options=options)

    def _fallback_log_normalization(self, norm, options):
        return znp.log(self._hook_normalization(norm=norm, options=options))

    @z.function(wraps="model")
    @deprecated_norm_range
    def ext_integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
        params: ztyping.ParamsTypeOpt = None,
    ) -> ztyping.XType:
        """Integrate the function over ``limits`` (normalized over ``norm_range`` if not False).

        Args:
            limits: |@doc:pdf.integrate.limits| Limits of the integration. |@docend:pdf.integrate.limits|
            norm: |@doc:pdf.integrate.norm| Normalization of the integration.
               By default, this is the same as the default space of the PDF.
               ``False`` means no normalization and returns the unnormed integral. |@docend:pdf.integrate.norm|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable,
                 non-dynamic and therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|
            params: |@doc:model.args.params| Mapping of the parameter names to the actual
               values. The parameter names refer to the names of the parameters,
               typically :py:class:`~zfit.Parameter`, that
               the model was _initialized_ with, not the name of the models
               parametrization. |@docend:model.args.params|

        Returns:
            The integral value as a scalar with shape ()
        """
        if options is None:
            options = {}
        norm = self._check_input_norm(norm)
        limits = self._check_input_limits(limits=limits)
        if not self.is_extended:
            msg = f"{self} is not extended, cannot call `ext_pdf`"
            raise NotExtendedPDFError(msg)
        with self._check_set_input_params(params=params):
            return self.integrate(limits=limits, norm=norm, options=options) * self.get_yield()

    def _apply_yield(self, value: float, norm: ztyping.LimitsType, log: bool) -> float | tf.Tensor:
        if self.is_extended and not norm.limits_are_false:
            if log:
                value += znp.log(self.get_yield())
            else:
                value *= self.get_yield()
        return value

    @deprecated(None, "Use the public `set_yield` instead.")
    def _set_yield_inplace(self, value: ZfitParameter | float | None):
        """Make the model extended by setting a yield.

        This does not alter the general behavior of the PDF. If there is a
        ``norm`` given, the output of the above functions does not represent a normalized
        probability density function anymore but corresponds to a number probability.

        Args:
            value:
        """
        self._set_yield(value=value)

    def create_extended(
        self,
        yield_: ztyping.ParamTypeInput,
        name: str | None = None,
        *,
        name_addition: str | None = None,
    ) -> ZfitPDF:
        """Return an extended version of this pdf with yield ``yield_``. The parameters are shared.

        Args:
            yield_: |@doc:pdf.param.yield| Yield (expected number of events) of the PDF.
               This is the expected number of events.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.param.yield|
            name: New name of the PDF. If ``None``, the name of the PDF with a trailing "_ext" is used.

        Returns:
            :py:class:`~zfit.core.interfaces.ZfitPDF`: a new PDF that is extended
        """
        # TODO(Mayou36): fix copy
        if name_addition is not None:
            msg = "name_addition is not supported anymore, use `name` instead."
            raise BreakingAPIChangeError(msg)
        from zfit.models.functor import ProductPDF

        name = f"{self.name}_ext" if name is None else name

        if isinstance(self, ProductPDF):
            warnings.warn(
                "As `copy` is not yet properly implemented, this may fails (for ProductPDF for example?). This"
                "will be fixed in the future.",
                category=UserWarning,
                stacklevel=2,
            )
        if self.is_extended:
            msg = "This PDF is already extended, cannot create an extended one."
            raise AlreadyExtendedPDFError(msg)
        try:
            new_pdf = self.copy(name=name)
        except Exception as error:
            msg = (
                f"PDF {self} could not be copied, therefore `create_extended` failed and a new "
                f"extended PDF cannot be created. As an alternative, you can use `set_yield`"
                f" to set the yield on the current PDF *inplace* (this won't return a new PDF but"
                f" instead modify the existing)."
            )
            raise RuntimeError(msg) from error
        new_pdf.set_yield(value=yield_)
        return new_pdf

    def set_yield(self, value):
        """Make the model extended **inplace** by setting a yield. If possible, prefer to use ``create_extended``.

        This does not alter the general behavior of the PDF. The ``pdf`` and ``integrate`` and similar methods will
        continue to return the same - normalized to 1 - values. However, not only can this parameter be accessed
        via ``get_yield``, the methods ``ext_pdf`` and ``ext_integral`` provide a version of ``pdf`` and ``integrate``
        respecetively that is multiplied by the yield.

        These can be useful for plotting and for binned likelihoods.

        Args:
            value: |@doc:pdf.param.yield| Yield (expected number of events) of the PDF.
               This is the expected number of events.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.param.yield|
        """
        self._set_yield(value=value)

    def _set_yield(self, value: ztyping.ParamTypeInput):
        if value is None:
            msg = "Cannot unset a yield (anymore)."
            raise BreakingAPIChangeError(msg)
        if self.is_extended:
            msg = f"Cannot extend {self}, is already extended."
            raise AlreadyExtendedPDFError(msg)
        value = convert_to_parameter(value)
        self.add_cache_deps(value)
        self._yield = value

        # not ideal, should be in parametrized. But we don't have too many base classes, so this should work
        self._assert_params_unique()

    @property
    def is_extended(self) -> bool:
        """Flag to tell whether the model is extended or not.

        Returns:
            A boolean.
        """
        return self._yield is not None

    def _hook_sample(self, limits, n):
        if n is None and self.is_extended:
            n = "extended"
        if isinstance(n, str) and n == "extended":
            if not self.is_extended:
                msg = "Cannot use 'extended' as value for `n` on a non-extended pdf."
                raise NotExtendedPDFError(msg)
            samples = extended_sampling(pdfs=self, limits=limits)
        elif isinstance(n, str):
            msg = "`n` is a string and not 'extended'. Other options are currently not implemented."
            raise ValueError(msg)
        elif n is None:
            msg = "`n` cannot be `None` if pdf is not extended."
            raise tf.errors.InvalidArgumentError(msg)
        else:
            samples = super()._hook_sample(limits=limits, n=n)
        return samples

    def get_yield(self) -> Parameter | None:
        """Return the yield (only for extended models).

        Returns:
            The yield of the current model or None
        """
        return self._yield

    @property
    def extended(self) -> Parameter | None:
        """Return the yield (only for extended models).

        Returns:
            The yield of the current model or None
        """
        return self.get_yield()

    def _get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
    ) -> set[ZfitParameter]:
        params = super()._get_params(floating, is_yield=is_yield, extract_independent=extract_independent)

        if is_yield is not False:
            if self.is_extended:
                yield_params = extract_filter_params(
                    self.get_yield(),
                    floating=floating,
                    extract_independent=extract_independent,
                )
                yield_params.update(params)  # putting the yields at the beginning
                params = yield_params
            elif is_yield is True:
                msg = "PDF is not extended but only yield parameters were requested."
                raise NotExtendedPDFError(msg)
        return params

    def create_projection_pdf(
        self,
        *,
        limits: ztyping.LimitsTypeInput = None,
        obs: ztyping.LimitsTypeInput = None,
        options=None,
        name: str | None = None,
        label: str | None = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
    ) -> ZfitPDF:
        """Create a PDF projection by integrating out some dimensions.

        The new projection pdf is still fully dependent on the pdf it was created with.

        Args:
            limits: Limits of the integration to project out. If not given, all observables that are not in `obs` are
                projected on using the default limits of the observables.
            obs: Observables to project on. If not given, all observables that are not in `limits` are projected on.
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable,
                 non-dynamic and therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|

        Returns:
            A pdf without the dimensions from ``limits``.
        """
        from ..models.special import SimpleFunctorPDF

        if limits is None:
            if obs is None:
                msg = "Either `limits` or `obs` have to be given."
                raise ValueError(msg)
            obs = convert_to_space(obs)
            limit_obs = [ob for ob in self.obs if ob not in obs.obs]
            if not limit_obs:
                msg = f"No observables to integrate out: `obs` contains all observables {obs}."
                raise ValueError(msg)
            limits = self.space.with_obs(limit_obs)
            if not obs.has_limits:
                obs = self.space.with_obs(obs.obs)
        else:
            limits = convert_to_space(limits)
            if not limits.has_limits:
                limits = self.space.with_obs(limits.obs)
            if obs is None:
                obs = self.space.with_obs([ob for ob in self.obs if ob not in limits.obs])
            else:
                obs = convert_to_space(obs)
                if not obs.has_limits:
                    obs = self.space.with_obs(obs.obs)
                if not set(obs.obs).isdisjoint(limits.obs):
                    msg = (
                        f"The `obs` to project on ({obs}) and the `limits` to integrate over ({limits}) "
                        "have to be disjoint."
                    )
                    raise ValueError(msg)

        def partial_integrate_wrapped(self_simple, x):
            del self_simple
            return self.partial_integrate(
                x, limits=limits, options=options, norm=False
            )  # todo: it should be fine not to normalize, right?

        if label is None:
            label = f"{self.label}_projon_{obs.obs[0]}"
        if extended is None:
            extended = self.is_extended
        if extended is True:
            extended = self.get_yield()
        return SimpleFunctorPDF(
            obs=obs,
            pdfs=(self,),
            func=partial_integrate_wrapped,
            name=name,
            label=label,
            extended=extended,
            norm=norm,
        )

    def copy(self, **override_parameters) -> BasePDF:
        """Creates a copy of the model.

        Note: the copy model may continue to depend on the original
        initialization arguments.

        Args:
          **override_parameters: String/value dictionary of initialization
            arguments to override with new value.

        Returns:
          A new instance of `type(self)` initialized from the union
            of self.parameters and override_parameters, i.e.,
            `dict(self.parameters, **override_parameters)`.
        """
        obs = self.norm

        # HACK(Mayou36): remove once copy is proper implemented
        from ..models.dist_tfp import WrapDistribution
        from ..models.kde import GaussianKDE1DimV1
        from ..models.polynomials import RecursivePolynomial

        if type(self) == WrapDistribution:  # NOT isinstance! Because e.g. Gauss wraps that and takes different args
            parameters = {"distribution": self._distribution, "dist_params": self.dist_params}
        else:
            # HACK END

            parameters = dict(self.params)
            lam = parameters.pop("lambda", None)
            if lam is not None:
                parameters["lam"] = lam

        if type(self) == GaussianKDE1DimV1:
            msg = (
                "Cannot copy `GaussianKDE1DimV1` (yet). If you tried to make it extended, use "
                "`set_yield`"
                " instead and set it inplace."
            )
            raise RuntimeError(msg)
            parameters["data"] = self._original_data

        # HACK(Mayou36): copy the polynomial correct, replace 'c_0' with coeff0/coeff_0 or similar
        if isinstance(self, RecursivePolynomial):
            parameters["coeff0"] = parameters.pop("c_0", None)
            coeffs = []
            i_coeff = 1
            # collect coeffs and convert to 'coeff' list
            while True:
                coeff_name = f"c_{i_coeff}"
                try:
                    coeff = parameters.pop(coeff_name)
                except KeyError:
                    break
                else:
                    coeffs.append(coeff)
                i_coeff += 1
            parameters["coeffs"] = coeffs

        from zfit.models.functor import BaseFunctor, SumPDF

        if isinstance(self, BaseFunctor):
            parameters = {}
            if isinstance(self, SumPDF):
                fracs = self.fracs
                if not self.is_extended:
                    fracs = fracs[:-1]
                parameters.update(fracs=fracs)
            parameters.update(pdfs=self.pdfs)

        parameters.update(obs=obs, name=self.name)
        parameters.update(**override_parameters)
        # if hasattr(self, "distribution"):
        #     parameters.update(distribution=self.distribution)
        yield_ = parameters.pop("yield", None)
        new_instance = type(self)(**parameters)
        if yield_ is not None:
            new_instance.set_yield(yield_)
        return new_instance

    @deprecated_norm_range
    def as_func(self, norm: ztyping.LimitsType = False):
        """Return a `Function` with the function `model(x, norm=norm)`.

        Args:
            norm: If not False or a `ZfitSpace`, this will be used to call the `pdf` function.
        """
        from .operations import convert_pdf_to_func  # prevent circular import

        return convert_pdf_to_func(pdf=self, norm=norm)

    def __str__(self):
        return f"{type(self).__name__} {self.label}"

    def to_unbinned(self):
        """Convert to unbinned pdf, returns self if already unbinned."""
        return self

    def to_binned(
        self,
        space: ztyping.SpaceType,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: Optional[str] = None,
        label: Optional[str] = None,
    ):
        """Convert to binned pdf, returns self if already binned."""
        from ..models.tobinned import BinnedFromUnbinnedPDF

        return BinnedFromUnbinnedPDF(pdf=self, space=space, extended=extended, norm=norm, name=name, label=label)

    def to_truncated(
        self,
        limits: ZfitSpace | Iterable[ZfitSpace] | None = None,
        *,
        obs=None,
        extended=None,
        norm=None,
        name: str | None = None,
        label: str | None = None,
    ):
        """Convert the PDF to a truncated version with possibly different and multiple limits.

        The arguments are the same as for :py:class:`~zfit.pdf.TruncatedPDF`, the only difference being that
        if no limits are given, the limit of the PDF is used, thereby truncating the PDF to its original limits.

        Args:
            pdf: The PDF to be truncated.
            limits: The limits to truncate the PDF. Can be a single limit or multiple limits.
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
               If None, the PDF will be extended if the original PDF is extended.
               If ``True`` and the original PDF is extended, the yield will be scaled to the
               fraction of the total integral that is within the limits.
               Therefore, the overall yield is comparable, i.e. the pdfs can be plotted
               "on top of each other".
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """

        from ..models.truncated import TruncatedPDF

        if limits is None:
            limits = obs if obs is not None else self.space
        if obs is None:
            obs = self.space
        if name is None:
            name = self.name + "_truncated"
        if label is None:
            label = self.label + " _truncated"
        if norm is None:
            norm = self.norm
        return TruncatedPDF(pdf=self, obs=obs, limits=limits, extended=extended, norm=norm, name=name, label=label)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest

import zfit


def test_get_dependents_is_deterministic():
    parameters = [zfit.Parameter(f"param{i}", i) for i in range(4)]
    obs = zfit.Space("obs1", (-3, 2))

    def create_pdf(params):
        param1, param2, param3, param4 = params
        gauss1 = zfit.pdf.Gauss(param1, param2, obs=obs)
        gauss2 = zfit.pdf.Gauss(param2, param2, obs=obs)
        gauss3 = zfit.pdf.Gauss(param2, param4, obs=obs)
        gauss4 = zfit.pdf.Gauss(param3, param1, obs=obs)
        sum_pdf = zfit.pdf.SumPDF(
            [gauss1, gauss2, gauss3, gauss4], fracs=[0.1, 0.3, 0.4]
        )
        return sum_pdf

    for params in (parameters, reversed(parameters)):
        pdf = create_pdf(params)
        assert (
            pdf.get_cache_deps() == pdf.get_cache_deps()
        ), "get_dependents() not deterministic"


def test_get_params():
    obs = zfit.Space("obs", (-4, 5))
    mu_nofloat = zfit.Parameter("mu_nofloat", 1, floating=False)
    mu2 = zfit.Parameter("mu2", 1)
    sigma2 = zfit.Parameter("sigma2", 2)
    sigma_comp = zfit.ComposedParameter("sigma_comp", lambda s: s * 0.7, params=sigma2)

    yield1 = zfit.Parameter("yield1", 10)
    yield2 = zfit.Parameter("yield2", 200)
    yield2_comp = zfit.ComposedParameter(
        "yield2_comp", lambda y: y * 0.9, params=yield2
    )

    gauss = zfit.pdf.Gauss(mu_nofloat, sigma_comp, obs)
    gauss2 = zfit.pdf.Gauss(mu2, sigma2, obs)
    gauss_ext = gauss.create_extended(yield1)
    gauss2_ext = gauss2.create_extended(yield2_comp)

    frac = zfit.Parameter("frac", 0.4)
    sum1 = zfit.pdf.SumPDF([gauss, gauss2], fracs=frac)
    sum1_ext = zfit.pdf.SumPDF([gauss_ext, gauss2_ext])

    assert set(gauss.get_params()) == {sigma2}

    with pytest.raises(ValueError):
        set(gauss.get_params(floating=True, is_yield=False, extract_independent=False))

    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=True, is_yield=False, extract_independent=False
            )
        )
    assert set(
        gauss_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu_nofloat, sigma_comp}
    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=True, is_yield=None, extract_independent=False
            )
        )

    assert set(
        gauss_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, yield1}
    with pytest.raises(ValueError):
        set(
            gauss_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    assert (
        set(
            gauss_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
        == set()
    )
    assert set(
        gauss_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {yield1}
    assert set(
        gauss_ext.get_params(floating=True, is_yield=True, extract_independent=False)
    ) == {yield1}

    # with extract deps
    assert set(
        gauss_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {sigma2}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu_nofloat, sigma2}
    assert set(
        gauss_ext.get_params(floating=False, is_yield=False, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        gauss_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {sigma2, yield1}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, sigma2, yield1}
    assert set(
        gauss_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        gauss_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield1}
    assert set(
        gauss_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield1}
    assert (
        set(
            gauss_ext.get_params(
                floating=False, is_yield=True, extract_independent=True
            )
        )
        == set()
    )

    # Gauss ext2
    assert set(
        gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=False)
    ) == {mu2, sigma2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu2, sigma2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )
        == set()
    )

    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=True, is_yield=None, extract_independent=False
            )
        )
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu2, sigma2, yield2_comp}
    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {yield2_comp}
    with pytest.raises(ValueError):
        set(
            gauss2_ext.get_params(
                floating=True, is_yield=True, extract_independent=False
            )
        )

    # with extract deps
    assert set(
        gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {mu2, sigma2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu2, sigma2}
    yield2.floating = False
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=True
            )
        )
        == set()
    )
    yield2.floating = True
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=False, extract_independent=True
            )
        )
        == set()
    )

    assert set(
        gauss2_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=None, extract_independent=True
            )
        )
        == set()
    )

    assert set(
        gauss2_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield2}
    assert set(
        gauss2_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield2}
    assert (
        set(
            gauss2_ext.get_params(
                floating=False, is_yield=True, extract_independent=True
            )
        )
        == set()
    )

    # sum extended
    with pytest.raises(ValueError):
        assert set(
            sum1_ext.get_params(
                floating=True, is_yield=False, extract_independent=False
            )
        )
    frac0 = sum1_ext.fracs[0]
    frac1 = sum1_ext.fracs[1]
    assert set(
        sum1_ext.get_params(floating=None, is_yield=False, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, mu2, sigma2, frac0, frac1}
    with pytest.raises(ValueError):
        assert set(
            sum1_ext.get_params(
                floating=False, is_yield=False, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(floating=True, is_yield=None, extract_independent=False)
        )
    assert set(
        sum1_ext.get_params(floating=None, is_yield=None, extract_independent=False)
    ) == {mu_nofloat, sigma_comp, mu2, sigma2, frac0, frac1, sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(
                floating=False, is_yield=None, extract_independent=False
            )
        )

    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(
                floating=False, is_yield=True, extract_independent=False
            )
        )
    assert set(
        sum1_ext.get_params(floating=None, is_yield=True, extract_independent=False)
    ) == {sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(
            sum1_ext.get_params(floating=True, is_yield=True, extract_independent=False)
        )

    # with extract deps
    assert set(
        sum1_ext.get_params(floating=True, is_yield=False, extract_independent=True)
    ) == {
        mu2,
        sigma2,
        yield1,
        yield2,  # fracs depend on them
    }
    assert set(
        sum1_ext.get_params(floating=None, is_yield=False, extract_independent=True)
    ) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=False, is_yield=False, extract_independent=True)
    ) == {mu_nofloat}

    assert set(
        sum1_ext.get_params(floating=True, is_yield=None, extract_independent=True)
    ) == {mu2, sigma2, yield1, yield2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=None, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat}
    yield1.floating = False
    assert set(
        sum1_ext.get_params(floating=False, is_yield=None, extract_independent=True)
    ) == {mu_nofloat, yield1}
    yield1.floating = True

    assert set(
        sum1_ext.get_params(floating=True, is_yield=True, extract_independent=True)
    ) == {yield1, yield2}
    assert set(
        sum1_ext.get_params(floating=None, is_yield=True, extract_independent=True)
    ) == {yield1, yield2}
    assert (
        set(
            sum1_ext.get_params(floating=False, is_yield=True, extract_independent=True)
        )
        == set()
    )


def test_unique_param_names():
    obs = zfit.Space("obs1", (-1, 2))

    param_a1 = zfit.Parameter("a", 1)
    param_a2 = zfit.Parameter("a", 2)
    param_a3 = zfit.Parameter("a", 5)
    param_a4fix = zfit.Parameter("a", 5, floating=False)
    param_b1 = zfit.Parameter("b", 7)
    param_b2 = zfit.Parameter("b", 3)
    param_c = zfit.Parameter("c", 0.42)

    compparam_a1 = zfit.ComposedParameter("a", lambda x: x, params=param_a1)
    compparam_a2 = zfit.ComposedParameter("a", lambda x: x, params=param_a2)

    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.pdf.Gauss(mu=param_a1, sigma=param_a2, obs=obs)

    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.pdf.Gauss(mu=param_a1, sigma=param_a4fix, obs=obs)

    gauss1 = zfit.pdf.Gauss(mu=param_a1, sigma=param_b1, obs=obs)
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        gauss1.create_extended(param_b2)
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        gauss1.set_yield(param_b2)

    gauss1 = zfit.pdf.Gauss(
        mu=param_a1, sigma=param_b1, obs=obs
    )  # recreate, it is now maybe broken
    gauss2 = zfit.pdf.Gauss(mu=param_a2, sigma=param_b1, obs=obs)
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.pdf.SumPDF([gauss1, gauss2], fracs=param_c)

    data = zfit.Data.from_numpy(obs=obs, array=np.random.normal(0, 1, size=100))
    gauss3 = zfit.pdf.Gauss(mu=param_a1, sigma=param_c, obs=obs)
    sumpdf = zfit.pdf.SumPDF([gauss1, gauss3], fracs=0.7)
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        sumpdf.create_extended(param_a3)

    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.loss.UnbinnedNLL(
            model=[sumpdf, gauss2], data=[data, data]
        )  # gauss1 and gauss2 don't go with each other

    sumpdfext = sumpdf.create_extended(param_b1)
    gaussext = gauss2.create_extended(15)
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.loss.ExtendedUnbinnedNLL(model=[sumpdfext, gaussext], data=[data, data])

    # composed parameters
    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.pdf.Gauss(mu=compparam_a1, sigma=param_a2, obs=obs)

    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.pdf.Gauss(mu=compparam_a1, sigma=param_a4fix, obs=obs)

    with pytest.raises(zfit.exception.ParamNameNotUniqueError):
        zfit.ComposedParameter("ue", lambda x: x, params=[param_a1, param_a2])



#  Copyright (c) 2024 zfit
from __future__ import annotations

import pickle

import zfit

n_bins = 50

# create space
obs = zfit.Space("x", -10, 10)

# parameters
mu = zfit.Parameter("mu", 1.0, -4, 6)
sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
n_bkg = zfit.Parameter("n_bkg", 20000)
n_sig = zfit.Parameter("n_sig", 1000)

# model building, pdf creation
gauss_extended = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs, extended=n_sig)
exp_extended = zfit.pdf.Exponential(lambd, obs=obs, extended=n_bkg)

model = zfit.pdf.SumPDF([gauss_extended, exp_extended])

data = model.create_sampler(n=21200)
data.resample()

# set the values to a start value for the fit
zfit.param.set_values({mu: 0.5, sigma: 1.2, lambd: -0.05})
# alternatively, we can set the values individually
# mu.set_value(0.5)
# sigma.set_value(1.2)
# lambd.set_value(-0.05)

# create NLL
nll = zfit.loss.ExtendedUnbinnedNLL(model=model, data=data)

# create a minimizer
# this uses the automatic gradient (more precise, but can fail)
minimizer = zfit.minimize.Minuit(gradient="zfit")
result = minimizer.minimize(nll).update_params()
# do the error calculations, here with hesse, than with minos
param_hesse = result.hesse()

# the second return value is a new FitResult if a new minimum was found

(
    param_errors,
    _,
) = result.errors()

# Storing the result can be achieved in many ways. Using dill (like pickle), we can dump and load the result
result_dilled = zfit.dill.dumps(result)
result_loaded = zfit.dill.loads(result_dilled)

# EXPERIMENTAL: we can serialize the model to a human-readable format with HS3
# human readable representation
hs3like = zfit.hs3.dumps(nll)
# print(hs3like)
# and we can load it again
nll_loaded = zfit.hs3.loads(hs3like)

# pickle can also be used, but as not all objects are pickleable, we need to freeze the result
# which makes it read-only and no new error calculations can be done
result.freeze()
dumped = pickle.dumps(result)
loaded = pickle.loads(dumped)

zfit.param.set_values(model.get_params(), loaded)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
from pytest_cases import parametrize, fixture

import zfit
import zfit.z.numpy as znp


# todo: expand tests with pdfs?

def pdf_gauss():
    scale = zfit.Parameter('scale1', 3065.)
    mu = zfit.Parameter('muparam', 1.2, -1., 3.)
    sigma = zfit.Parameter('sigmaparam', 13, 0.1, 100.)
    obs = zfit.Space('obs1', -10, 10)

    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs, extended=scale)
    return gauss


def pdf_gaussnonext():
    mu = zfit.Parameter('muparam', 1.2, -1., 3.)
    sigma = zfit.Parameter('sigmaparam', 1.3, 0.1, 10.)
    obs = zfit.Space('obs1', -10, 10)

    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    return gauss


def pdf_expnonext():
    lambda_ = zfit.Parameter('lambda', -0.03, -2., 0.)
    obs = zfit.Space('obs1', -10, 10)

    exp = zfit.pdf.Exponential(lambda_=lambda_, obs=obs)
    return exp


def pdf_sumnonext():
    sum_pdf = zfit.pdf.SumPDF(pdfs=[pdf_gaussnonext(), pdf_expnonext()], fracs=zfit.Parameter('frac', 0.25))
    return sum_pdf


def pdf_exp():
    lambda_ = zfit.Parameter('lambda', -0.03, -2., 0.)
    scale = zfit.Parameter('scale2', 10001.)
    obs = zfit.Space('obs1', -10, 10)

    exp = zfit.pdf.Exponential(lambda_=lambda_, obs=obs, extended=scale)
    return exp


def pdf_cheby():
    c0 = zfit.Parameter('c0', 1.)
    c1 = zfit.Parameter('c1', 0.5)
    obs = zfit.Space('obs1', -10, 10)
    scale = zfit.Parameter('scale3', 10101.)

    cheby = zfit.pdf.Chebyshev(obs=obs, coeffs=[c0, c1], extended=scale)
    return cheby


def pdf_prod():
    prod = zfit.pdf.ProductPDF(pdfs=[pdf_gauss(), pdf_cheby()], extended=pdf_gauss().get_yield())
    return prod


def pdf_sum():
    sum_pdf = zfit.pdf.SumPDF(pdfs=[pdf_prod(), pdf_exp()])
    return sum_pdf


# @parametrize('pdf', [pdf_gauss, pdf_exp, pdf_cheby, pdf_prod, pdf_sum])
@pytest.mark.parametrize('binning', [None, 113])
def test_pdf_params_args_methods(binning):
    mu = zfit.Parameter('muparam', 1.2)
    sigma = zfit.Parameter('sigmaparam', 1.1)
    scale = zfit.Parameter('scale', 10.)
    obs = zfit.Space('obs1', -10, 10, binning=binning)

    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs, extended=scale)
    if binning is not None:
        gauss = gauss.to_binned(obs)

    data = znp.array([1.1, 1.2, 1.3, 1.4, 1.5])

    assert np.argmax(gauss.pdf(data)) == 1
    probs = gauss.pdf(data, params={'muparam': 1.4})
    assert np.argmax(probs) == 3
    with pytest.raises(ValueError):
        gauss.pdf(data, params={'muparam': 1.4, 'sigma': 1.5})
    assert np.argmax(gauss.pdf(data)) == 1

    assert np.argmax(gauss.ext_pdf(data, params={'muparam': 1.4, "sigmaparam": 1.1})) == 3
    with pytest.raises(ValueError):
        gauss.ext_pdf(data, params={'mupar': 1.4, 'sigmaparam': 1.5})
    assert np.argmax(gauss.ext_pdf(data)) == 1

    assert np.argmax(gauss.log_pdf(data, params={'muparam': 1.4, "sigmaparam": 1.1})) == 3
    with pytest.raises(ValueError):
        gauss.log_pdf(data, params={'mupar': 1.4, 'sigmaparam': 1.5})
    assert np.argmax(gauss.log_pdf(data)) == 1

    assert np.argmax(gauss.ext_log_pdf(data, params={'muparam': 1.4, "sigmaparam": 1.1})) == 3
    with pytest.raises(ValueError):
        gauss.ext_log_pdf(data, params={'mupar': 1.4, 'sigmaparam': 1.5})

    assert np.argmax(gauss.ext_log_pdf(data)) == 1

    integral_diff = gauss.integrate(limits=(-1, 2)) - gauss.integrate(limits=(-1, 2), params={'sigmaparam': 1.4})
    assert integral_diff > 0
    with pytest.raises(ValueError):
        gauss.integrate(limits=(-1, 2), params={'sigmaparam': 1.4, 'mu': 1.5})

    integral_diff = gauss.ext_integrate(limits=(-1, 2)) - gauss.ext_integrate(limits=(-1, 2), params={'sigmaparam': 1.4})
    assert integral_diff > 0
    with pytest.raises(ValueError):
        gauss.ext_integrate(limits=(-1, 2), params={'sigmaparam': 1.4, 'mu': 1.5})

    if binning is not None:
        assert np.argmax(gauss.counts(data)) == 1
        assert np.argmax(gauss.counts(data, params={'muparam': 1.4})) == 3
        with pytest.raises(ValueError):
            gauss.counts(data, params={'muparam': 1.4, 'sigma': 1.5})
        assert np.argmax(gauss.counts(data)) == 1

        assert np.argmax(gauss.rel_counts(data, params={'muparam': 1.4, "sigmaparam": 1.1})) == 3
        with pytest.raises(ValueError):
            gauss.rel_counts(data, params={'mupar': 1.4, 'sigmaparam': 1.5})
        assert np.argmax(gauss.rel_counts(data)) == 1


def pdf_binned():
    return pdf_sum().to_binned(142)


def unbinnednll():
    pdf = pdf_sumnonext()
    data = pdf.create_sampler(n=10000)
    loss = zfit.loss.UnbinnedNLL(model=pdf, data=data)
    return loss


def ext_unbinnednll():
    pdf = pdf_sum()
    data = pdf.create_sampler(n=10000)
    loss = zfit.loss.ExtendedUnbinnedNLL(model=pdf, data=data)
    return loss


def binnednll():
    pdf = pdf_binned()
    databinned = pdf.sample(n=10000)
    loss = zfit.loss.BinnedNLL(model=pdf, data=databinned)
    return loss


def ext_binnednll():
    pdf = pdf_binned()
    databinned = pdf.sample(n=10000)
    loss = zfit.loss.ExtendedBinnedNLL(model=pdf, data=databinned)
    return loss


def chi2():
    pdf = pdf_binned()
    databinned = pdf.sample(n=10000)
    databinned = databinned.with_variances(databinned.counts())
    loss = zfit.loss.BinnedChi2(model=pdf, data=databinned)
    return loss


def ext_chi2():
    pdf = pdf_binned()
    databinned = pdf.sample(n=10000)
    databinned = databinned.with_variances(databinned.counts())
    loss = zfit.loss.ExtendedBinnedChi2(model=pdf, data=databinned)
    return loss


losses = [unbinnednll, ext_unbinnednll, binnednll, ext_binnednll, chi2, ext_chi2]


@pytest.mark.parametrize('loss', losses)
def test_params_as_args_loss(loss):
    loss = loss()
    params = loss.get_params()
    param1 = list(params)[0]
    loss_values = loss.value()
    assert np.isfinite(loss_values)
    loss_values_params = loss.value(params={param1.name: param1 * 1.1})
    assert abs(loss_values - loss_values_params) > 0.1

@pytest.mark.parametrize('loss', losses)
@pytest.mark.flaky(reruns=2)  # minimization not reliable
def test_params_as_args_fitresult(loss):
    loss = loss()
    minimizer = zfit.minimize.Minuit()
    result = minimizer.minimize(loss)
    assert result.valid, "Test is flawed"
    params = loss.get_params(is_yield=False)
    param1 = list(params)[0]
    loss_values = result.fmin
    assert np.isfinite(loss_values)
    param1min = param1.value()
    param1.set_value(param1 *0.1)
    loss_values_nonmin = loss.value()
    assert np.isfinite(loss_values_nonmin), "Test is flawed, loss is not finite"
    assert loss_values_nonmin - loss_values > 0.1
    loss_values_params = loss.value(params={param1.name: param1min})
    assert pytest.approx(loss_values) == loss_values_params
    loss_values_result = loss.value(params=result)
    assert pytest.approx(loss_values) == loss_values_result



"""This module contains functions for the numeric as well as the analytic (partial) integration."""

#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Mapping, Optional

if TYPE_CHECKING:
    pass

import collections
from collections.abc import Callable
from contextlib import suppress

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

import zfit.z.numpy as znp
from zfit import z

from ..settings import ztypes
from ..util import ztyping
from ..util.exception import AnalyticIntegralNotImplemented, WorkInProgressError
from .interfaces import ZfitModel, ZfitSpace
from .space import MultiSpace, convert_to_space, supports


def auto_integrate(
    func,
    limits,
    n_axes=None,
    x=None,
    method="AUTO",
    dtype=ztypes.float,
    mc_sampler=tfp.mcmc.sample_halton_sequence,
    max_draws=None,
    tol=None,
    vectorizable=None,
    mc_options=None,
    simpsons_options=None,
):
    if vectorizable is None:
        vectorizable = False
    limits = convert_to_space(limits)

    if n_axes is None:
        n_axes = limits.n_obs
    if method == "AUTO":  # TODO unfinished, other methods?
        method = "simpson" if n_axes == 1 and x is None else "mc"
    # TODO method
    if method.lower() == "mc":
        mc_options = mc_options or {}
        draws_per_dim = mc_options["draws_per_dim"]
        max_draws = mc_options.get("max_draws")
        integral = mc_integrate(
            x=x,
            func=func,
            limits=limits,
            n_axes=n_axes,
            method=method,
            dtype=dtype,
            mc_sampler=mc_sampler,
            draws_per_dim=draws_per_dim,
            max_draws=max_draws,
            tol=tol,
            importance_sampling=None,
            vectorizable=vectorizable,
        )
    elif method.lower() == "simpson":
        num_points = simpsons_options["draws_simpson"]
        integral = simpson_integrate(func=func, limits=limits, num_points=num_points)
    else:
        msg = f"Method {method} not a legal choice for integration method."
        raise ValueError(msg)
    return znp.atleast_1d(integral)


# TODO implement numerical integration method
def numeric_integrate():
    """Integrate ``func`` using numerical methods."""
    return


# COPIED FROM tf_quant_finance/math/integrate.py which is deprecated
def simpson(func, lower, upper, num_points=1001, dtype=None):
    """Evaluates definite integral using composite Simpson's 1/3 rule.

    Integrates `func` using composite Simpson's 1/3 rule [1].

    Evaluates function at points of evenly spaced grid of `num_points` points,
    then uses obtained values to interpolate `func` with quadratic polynomials
    and integrates these polynomials.

    #### References
    [1] Weisstein, Eric W. "Simpson's Rule." From MathWorld - A Wolfram Web
        Resource. http://mathworld.wolfram.com/SimpsonsRule.html

    #### Example
    ```python
      f = lambda x: x*x
      a = tf.constant(0.0)
      b = tf.constant(3.0)
      integrate(f, a, b, num_points=1001) # 9.0
    ```

    Args:
      func: Python callable representing a function to be integrated. It must be a
        callable of a single `Tensor` parameter and return a `Tensor` of the same
        shape and dtype as its input. It will be called with a `Tesnor` of shape
        `lower.shape + [n]` (where n is integer number of points) and of the same
        `dtype` as `lower`.
      lower: `Tensor` or Python float representing the lower limits of
        integration. `func` will be integrated between each pair of points defined
        by `lower` and `upper`.
      upper: `Tensor` of the same shape and dtype as `lower` or Python float
        representing the upper limits of intergation.
      num_points: Scalar int32 `Tensor`. Number of points at which function `func`
        will be evaluated. Must be odd and at least 3.
        Default value: 1001.
      dtype: Optional `tf.Dtype`. If supplied, the dtype for the `lower` and
        `upper`. Result will have the same dtype.
        Default value: None which maps to dtype of `lower`.
      name: Python str. The name to give to the ops created by this function.
        Default value: None which maps to 'integrate_simpson_composite'.

    Returns:
      `Tensor` of shape `func_batch_shape + limits_batch_shape`, containing
        value of the definite integral.
    """
    lower = tf.convert_to_tensor(lower, dtype=dtype)
    dtype = lower.dtype
    upper = tf.convert_to_tensor(upper, dtype=dtype)
    num_points = tf.convert_to_tensor(num_points, dtype=tf.int32, name="num_points")

    assertions = [
        tf.debugging.assert_greater_equal(num_points, 3),
        tf.debugging.assert_equal(num_points % 2, 1),
    ]

    with tf.control_dependencies(assertions):
        dx = (upper - lower) / (znp.asarray(num_points, dtype=dtype) - 1)
        dx_expand = tf.expand_dims(dx, -1)
        lower_exp = tf.expand_dims(lower, -1)
        grid = lower_exp + dx_expand * znp.asarray(tf.range(num_points), dtype=dtype)
        weights_first = tf.constant([1.0], dtype=dtype)
        weights_mid = tf.tile(tf.constant([4.0, 2.0], dtype=dtype), [(num_points - 3) // 2])
        weights_last = tf.constant([4.0, 1.0], dtype=dtype)
        weights = tf.concat([weights_first, weights_mid, weights_last], axis=0)

    return znp.sum(func(grid) * weights, axis=-1) * dx / 3.0


def simpson_integrate(func, limits, num_points):  # currently not vectorized
    integrals = []
    num_points = znp.asarray(num_points, znp.int32)
    num_points += num_points % 2 + 1  # sanitize number of points
    for space in limits:
        lower, upper = space.v0.limits  # todo: shape of spaces?
        if lower.shape[0] > 1:
            msg = "Vectorized spaces in integration currently not supported."
            raise ValueError(msg)
        lower = znp.array(lower)[0, 0]
        upper = znp.array(upper)[0, 0]
        tf.debugging.assert_all_finite(
            (lower, upper),
            "MC integration does (currently) not support unbound limits (np.infty) as given here:"
            f"\nlower: {lower}, upper: {upper}",
        )
        integrals.append(
            simpson(
                func=func,
                lower=lower,
                upper=upper,
                num_points=num_points,
                dtype=znp.float64,
            )
        )
    return znp.sum(integrals, axis=0)


# @z.function
def mc_integrate(
    func: Callable,
    limits: ztyping.LimitsType,
    axes: ztyping.AxesTypeInput | None = None,
    x: ztyping.XType | None = None,
    n_axes: int | None = None,
    draws_per_dim: int = 40000,
    max_draws=800_000,
    tol: float = 1e-6,
    method: Optional[str] = None,  # noqa: ARG001
    xfixed: Mapping | None = None,
    dtype: type = ztypes.float,
    mc_sampler: Callable = tfp.mcmc.sample_halton_sequence,  # noqa: ARG001
    importance_sampling: Callable | None = None,  # noqa: ARG001
    vectorizable=None,
) -> tf.Tensor:
    """Monte Carlo integration of ``func`` over ``limits``.

    Args:
        vectorizable ():
        func: The function to be integrated over
        limits: The limits of the integral
        axes: The row to integrate over. None means integration over all value
        x: If a partial integration is performed, this are the value where x will be evaluated.
        n_axes: the number of total dimensions (old?)
        draws_per_dim: How many random points to draw per dimensions
        method: Which integration method to use
        dtype: |dtype_arg_descr|
        mc_sampler: A function that takes one argument (``n_draws`` or similar) and returns
            random value between 0 and 1.
        importance_sampling:

    Returns:
        The integral
    """
    # if method is not None:
    #     raise ValueError("Method is not yet implemented.")
    # if importance_sampling is not None:
    #     raise ValueError("Importance sampling is not yet implemented.")

    import zfit

    if vectorizable is None:
        vectorizable = False
    tol = znp.array(tol, dtype=znp.float64)
    if axes is not None and n_axes is not None:
        msg = "Either specify axes or n_axes"
        raise ValueError(msg)

    axes = limits.axes
    partial = (axes is not None) and (x is not None)  # axes, value can be tensors

    if axes is not None and n_axes is None:
        n_axes = len(axes)
    if n_axes is not None and axes is None:
        axes = tuple(range(n_axes))

    if partial:
        if isinstance(limits, MultiSpace):
            msg = "Partial integration not supported with multiple limits."
            raise ValueError(msg)
        space = limits
        data_obs = x.obs
        integrate_data = x.value()

        # todo: more finegrained control over vectorize vs map_fn?
        def map_eager(func: Callable, iterable: Iterable):
            return znp.asarray(tuple(map(func, iterable)))

        mapfn = tf.vectorized_map if (vectorize_partial := zfit.run.get_graph_mode() is not False) else map_eager

        if xfixed is not None:
            msg = "xfixed should be None"
            raise RuntimeError(msg)

        @z.function(wraps="tensor")
        def part_integrate_func(x):  # TODO: improve? as_tensormap or similar?
            x = dict(zip(data_obs, tf.unstack(x, axis=0)))
            return mc_integrate(
                func=func,
                limits=limits,
                axes=axes,
                x=None,
                xfixed=x,
                vectorizable=vectorize_partial,
            )

        return mapfn(
            part_integrate_func,
            integrate_data,
        )[:, 0]

    integrals = []
    for space in limits:
        lower, upper = space.v1.limits
        # tf.debugging.assert_all_finite(
        #     (lower, upper),
        #     "MC integration does (currently) not support unbound limits (np.infty) as given here:"
        #     f"\nlower: {lower}, upper: {upper}",
        # )

        n_samples = draws_per_dim * n_axes

        chunked_normalization = zfit.run.chunksize < n_samples
        if chunked_normalization and partial:
            pass
        if chunked_normalization and not partial:
            n_chunks = int(np.ceil(n_samples / zfit.run.chunksize))
            chunksize = int(np.ceil(n_samples / n_chunks))
            avg = normalization_chunked(
                func=func,
                n_axes=n_axes,
                dtype=dtype,
                x=x,
                num_batches=n_chunks,
                batch_size=chunksize,
                space=space,
            )

        else:
            # TODO: deal with n_obs properly?
            @z.function(wraps="tensor", keepalive=True)
            def cond(avg, error, std, ntot, i):
                del avg, std, i
                return znp.logical_and(error > tol, ntot < max_draws)

            # binding the variables to the function
            @z.function(wraps="tensor")
            def body_integrate(avg, error, std, ntot, i, n_samples=n_samples, upper=upper, lower=lower, space=space):
                ntot_old = ntot
                ntot += n_samples
                if partial:
                    samples_normed = tfp.mcmc.sample_halton_sequence(
                        dim=n_axes,
                        # sequence_indices=tf.range(ntot_old, ntot),
                        num_results=n_samples / 10,
                        # reduce, it explodes otherwise easily
                        # as we don't do adaptive now
                        # to decrease integration size
                        dtype=dtype,
                        randomized=False,
                    )
                else:
                    samples_normed = tfp.mcmc.sample_halton_sequence(
                        dim=n_axes,
                        sequence_indices=tf.range(ntot_old, ntot),
                        # num_results=n_samples,  # to decrease integration size
                        dtype=dtype,
                        randomized=False,
                    )
                samples = samples_normed * (upper - lower) + lower  # samples is [0, 1], stretch it

                if xfixed is not None:
                    shape = tf.shape(samples)[0]
                    xfixed_shaped = {key: znp.broadcast_to(val, shape) for key, val in xfixed.items()}
                    samples_named = dict(zip(space.obs, tf.unstack(samples, axis=1)))
                    samples_tot = {**samples_named, **xfixed_shaped}
                    obstot, samplestot = zip(*samples_tot.items())
                    samplestot = tf.stack(samplestot, axis=-1)
                    samples = zfit.Data.from_tensor(tensor=samplestot, obs=obstot)

                xval = samples
                y = func(xval)
                ifloat = znp.asarray(i, dtype=tf.float64)
                avg = avg / (ifloat + 1.0) * ifloat + znp.mean(y) / (ifloat + 1.0)
                std = std / (ifloat + 1.0) * ifloat + znp.std(y) / (ifloat + 1.0)
                ntot_float = znp.asarray(ntot, dtype=znp.float64)

                # estimating the error of QMC is non-trivial
                # (https://www.degruyter.com/document/doi/10.1515/mcma-2020-2067/html or
                # https://stats.stackexchange.com/questions/533725/how-to-calculate-quasi-monte-carlo-integration-error-when-sampling-with-sobols)
                # However, we use here just something that is in the right direction for the moment being. Therefore,
                # we use the MC error as an upper bound as QMC is better/equal to MC (for our cases).
                error_sobol = std * znp.log(ntot_float) ** n_axes / ntot_float
                error_random = std / znp.sqrt(ntot_float)
                error = znp.minimum(error_sobol, error_random) * 0.1  # heuristic factor from using QMC
                return avg, error, std, ntot, i + 1

            avg, error, std, ntot, i = [
                znp.array(0.0, dtype=znp.float64),
                znp.array(9999.0, dtype=znp.float64),  # init value large, irrelevant
                znp.array(0.0, dtype=znp.float64),
                znp.array(0.0, dtype=znp.int64),
                znp.array(0.0, dtype=znp.int64),
            ]
            if vectorizable:
                avg, error, std, ntot, i = body_integrate(avg, error, std, ntot, i)
            else:
                avg, error, std, ntot, i = tf.while_loop(
                    cond=cond, body=body_integrate, loop_vars=[avg, error, std, ntot, i]
                )
                from zfit import settings

                if settings.get_verbosity() > 9:
                    tf.print("i:", i, "   ntot:", ntot)

            def print_none_return(error=error):
                from zfit import settings

                if settings.get_verbosity() >= 0:
                    tf.print(
                        "Estimated integral error (",
                        error,
                        ") larger than tolerance (",
                        tol,
                        "), which is maybe not enough (but maybe it's also fine)."
                        " You can (best solution) implement an anatytical integral (see examples in repo) or"
                        " manually set a higher number on the PDF with 'update_integration_options'"
                        " and increase the 'max_draws' (or adjust 'tol'). "
                        "If partial integration is chosen, this can lead to large memory consumption."
                        "This is a new warning checking the integral accuracy. It may warns too often as it is"
                        " Work In Progress. If you have any observation on it, please tell us about it:"
                        " https://github.com/zfit/zfit/issues/new/choose"
                        "To suppress this warning, use zfit.settings.set_verbosity(-1).",
                    )

            if not vectorizable:
                tf.cond(error > tol, print_none_return, lambda: None)
        integral = avg * znp.asarray(z.convert_to_tensor(space.area()), dtype=avg.dtype)
        integrals.append(integral)
    integral = z.reduce_sum(integrals, axis=0)
    return znp.atleast_1d(integral)


# TODO(Mayou36): Make more flexible for sampling
# @z.function
def normalization_nograd(func, n_axes, batch_size, num_batches, dtype, space, x=None, shape_after=()):
    del x  # unused
    upper, lower = space.v1.limits
    lower = z.convert_to_tensor(lower, dtype=dtype)
    upper = z.convert_to_tensor(upper, dtype=dtype)

    def body(batch_num, mean):
        start_idx = batch_num * batch_size
        end_idx = start_idx + batch_size
        indices = tf.range(start_idx, end_idx, dtype=tf.int32)
        samples_normed = tfp.mcmc.sample_halton_sequence(
            n_axes,
            # num_results=batch_size,
            sequence_indices=indices,
            dtype=dtype,
            randomized=False,
        )
        # halton_sample = tf.random_uniform(shape=(n_axes, batch_size), dtype=dtype)
        samples_normed.set_shape((batch_size, n_axes))
        samples_normed = znp.expand_dims(samples_normed, axis=0)
        samples = samples_normed * (upper - lower) + lower
        func_vals = func(samples)
        if shape_after == ():
            reduce_axis = None
        else:
            reduce_axis = 1
            if len(func_vals.shape) == 1:
                func_vals = znp.expand_dims(func_vals, -1)
        batch_mean = znp.mean(func_vals, axis=reduce_axis)  # if there are gradients
        err_weight = 1 / znp.asarray(batch_num + 1, dtype=tf.float64)

        do_print = False
        if do_print:
            tf.print(batch_num + 1)
        return batch_num + 1, mean + err_weight * (batch_mean - mean)

    def cond(batch_num, _):
        return batch_num < num_batches

    initial_mean = tf.constant(0, shape=shape_after, dtype=dtype)
    initial_body_args = (0, initial_mean)
    _, final_mean = tf.while_loop(
        cond=cond,
        body=body,
        loop_vars=initial_body_args,
        parallel_iterations=1,
        swap_memory=False,
        back_prop=True,
    )
    # def normalization_grad(x):
    return final_mean


# @z.function
def normalization_chunked(func, n_axes, batch_size, num_batches, dtype, space, x=None, shape_after=()):
    x_is_none = x is None

    @tf.custom_gradient
    def normalization_func(x):
        if x_is_none:
            x = None
        value = normalization_nograd(
            func=func,
            n_axes=n_axes,
            batch_size=batch_size,
            num_batches=num_batches,
            dtype=dtype,
            space=space,
            x=x,
            shape_after=shape_after,
        )

        def grad_fn(dy, variables=None):
            if variables is None:
                return dy, None
            with tf.GradientTape() as tape:
                value = normalization_nograd(
                    func=func,
                    n_axes=n_axes,
                    batch_size=batch_size,
                    num_batches=num_batches,
                    dtype=dtype,
                    space=space,
                    x=x,
                    shape_after=shape_after,
                )

            return dy, tape.gradient(value, variables)

        return value, grad_fn

    fake_x = 1 if x_is_none else x
    return normalization_func(fake_x)


# @z.function
def chunked_average(func, x, num_batches, batch_size, space, mc_sampler):
    lower, upper = space.v0.limits

    fake_resource_var = tf.Variable("fake_hack_ResVar_for_custom_gradient", initializer=z.constant(4242.0))
    fake_x = z.constant(42.0) * fake_resource_var

    @tf.custom_gradient
    def dummy_func(fake_x):  # to make working with custom_gradient  # noqa: ARG001
        if x is not None:
            msg = "partial not yet implemented"
            raise WorkInProgressError(msg)

        def body(batch_num, mean):
            if mc_sampler == tfp.mcmc.sample_halton_sequence:
                start_idx = batch_num * batch_size
                end_idx = start_idx + batch_size
                indices = tf.range(start_idx, end_idx, dtype=tf.int32)
                sample = mc_sampler(
                    space.n_obs,
                    sequence_indices=indices,
                    dtype=ztypes.float,
                    randomized=False,
                )
            else:
                sample = mc_sampler(shape=(batch_size, space.n_obs), dtype=ztypes.float)
            sample = tf.guarantee_const(sample)
            sample = (np.array(upper[0]) - np.array(lower[0])) * sample + lower[0]
            sample = znp.transpose(a=sample)
            sample = func(sample)
            sample = tf.guarantee_const(sample)

            batch_mean = znp.mean(sample)
            batch_mean = tf.guarantee_const(batch_mean)
            err_weight = 1 / znp.asarray(batch_num + 1, dtype=tf.float64)
            # err_weight /= err_weight + 1
            # print_op = tf.print(batch_mean)
            do_print = False
            if do_print:
                tf.print(batch_num + 1, mean, err_weight * (batch_mean - mean))

            return batch_num + 1, mean + err_weight * (batch_mean - mean)

        def cond(batch_num, _):
            return batch_num < num_batches

        initial_mean = tf.convert_to_tensor(value=0, dtype=ztypes.float)
        _, final_mean = tf.while_loop(
            cond=cond,
            body=body,
            loop_vars=(0, initial_mean),
            parallel_iterations=1,
            swap_memory=False,
            back_prop=False,
            maximum_iterations=num_batches,
        )

        def dummy_grad_with_var(dy, variables=None):
            msg = "Who called me? Mayou36"
            raise WorkInProgressError(msg)
            if variables is None:
                msg = "Is this needed? Why? It's not a NN. Please make an issue."
                raise WorkInProgressError(msg)

            def dummy_grad_func(x):
                values = func(x)
                return tf.gradients(ys=values, xs=variables, grad_ys=dy) if variables else None

            return chunked_average(
                func=dummy_grad_func,
                x=x,
                num_batches=num_batches,
                batch_size=batch_size,
                space=space,
                mc_sampler=mc_sampler,
            )

        def dummy_grad_without_var(dy):
            return dummy_grad_with_var(dy=dy, variables=None)

        do_print = False
        if do_print:
            tf.print("Total mean calculated = ", final_mean)

        return final_mean, dummy_grad_with_var

    try:
        return dummy_func(fake_x)
    except TypeError:
        return dummy_func(fake_x)


class AnalyticIntegral:
    def __init__(self, *args, **kwargs):
        """Hold analytic integrals and manage their dimensions, limits etc."""
        super().__init__(*args, **kwargs)
        self._integrals = collections.defaultdict(dict)

    def get_max_axes(self, limits: ztyping.LimitsType) -> tuple[int]:
        """Return the maximal available axes to integrate over analytically for given limits.

        Args:
            limits: The integral function will be able to integrate over this limits
            axes: The axes over which (or over a subset) it will integrate

        Returns:
            Tuple[int]:
        """
        if not isinstance(limits, ZfitSpace):
            msg = "`limits` have to be a `ZfitSpace`"
            raise TypeError(msg)
        # limits = convert_to_space(limits=limits)

        return self._get_max_axes_limits(limits, out_of_axes=limits.axes)[0]  # only axes

    def _get_max_axes_limits(self, limits, out_of_axes):  # TODO: automatic caching? but most probably not relevant
        if out_of_axes:
            out_of_axes = frozenset(out_of_axes)
            implemented_axes = frozenset(d for d in self._integrals if d <= out_of_axes)
        else:
            implemented_axes = set(self._integrals.keys())
        implemented_axes = sorted(implemented_axes, key=len, reverse=True)  # iter through biggest first
        for axes in implemented_axes:
            limits_matched = [lim for lim, integ in self._integrals[axes].items() if integ.limits >= limits]

            if limits_matched:  # one or more integrals available
                return tuple(sorted(axes)), limits_matched
        return (), ()  # no integral available for this axes

    def get_max_integral(self, limits: ztyping.LimitsType, axes: ztyping.AxesTypeInput = None) -> None | Integral:
        """Return the integral over the ``limits`` with ``axes`` (or a subset of them).

        Args:
            limits:
            axes:

        Returns:
            Return a callable that integrated over the given limits.
        """
        limits = convert_to_space(limits=limits, axes=axes)

        axes, limits = self._get_max_axes_limits(limits=limits, out_of_axes=axes)
        axes = frozenset(axes)
        integrals = [self._integrals[axes][lim] for lim in limits]
        return max(integrals, key=lambda lim: lim.priority, default=None)

    def register(
        self,
        func: Callable,
        limits: ztyping.LimitsType,
        priority: int = 50,
        *,
        supports_norm: bool = False,
        supports_multiple_limits: bool = False,
    ) -> None:
        """Register an analytic integral.

        Args:
            func: The integral function. Takes 1 argument.
            axes: |dims_arg_descr|
            limits: |limits_arg_descr| ``Limits`` can be None if ``func`` works for any
            possible limits
            priority: If two or more integrals can integrate over certain limits, the one with the higher
                priority is taken (usually around 0-100).
            supports_norm: If True, norm_range will (if needed) be given to ``func`` as an argument.
            supports_multiple_limits: If True, multiple limits may be given as an argument to ``func``.
        """

        # if limits is False:
        #     raise ValueError("Limits for the analytical integral have to be specified or None (for any limits).")
        # if limits is None:
        #     limits = tuple((Space.ANY_LOWER, Space.ANY_UPPER) for _ in range(len(axes)))
        #     limits = convert_to_space(axes=axes, limits=limits)
        # else:
        #     limits = convert_to_space(axes=self.axes, limits=limits)
        # limits = limits.get_limits()
        if not isinstance(limits, ZfitSpace):
            msg = "Limits for registering an integral have to be `ZfitSpace`"
            raise TypeError(msg)
        axes = frozenset(limits.axes)

        # add catching everything unsupported:
        func = supports(norm=supports_norm, multiple_limits=supports_multiple_limits)(func)
        limits = limits.with_axes(axes=tuple(sorted(limits.axes)))
        self._integrals[axes][limits] = Integral(func=func, limits=limits, priority=priority)  # TODO improve with
        # database-like access

    def integrate(
        self,
        x: ztyping.XType | None,
        limits: ztyping.LimitsType,
        axes: ztyping.AxesTypeInput = None,
        norm: ztyping.LimitsType = None,
        model: ZfitModel = None,
        params: dict | None = None,
    ) -> ztyping.XType:
        """Integrate analytically over the axes if available.

        Args:
            x: If a partial integration is made, x are the value to be evaluated for the partial
                integrated function. If a full integration is performed, this should be `None`.
            limits: The limits to integrate
            axes: The dimensions to integrate over
            norm: |norm_range_arg_descr|
            params: The parameters of the function


        Returns:
            Union[tf.Tensor, float]:

        Raises:
            AnalyticIntegralNotImplementedError: If the requested integral is not available.
        """
        if axes is None:
            axes = limits.axes
        axes = frozenset(axes)
        integral_holder = self._integrals.get(axes)
        # limits = convert_to_space(axes=self.axes, limits=limits)
        if integral_holder is None:
            msg = f"Analytic integral is not available for axes {axes}"
            raise AnalyticIntegralNotImplemented(msg)
        integral_fn = self.get_max_integral(limits=limits)
        if integral_fn is None:
            msg = f"Integral is available for axes {axes}, but not for limits {limits}"
            raise AnalyticIntegralNotImplemented(msg)

        with suppress(TypeError):
            return integral_fn(x=x, limits=limits, norm=norm, params=params, model=model)
        with suppress(TypeError):
            return integral_fn(limits=limits, norm=norm, params=params, model=model)

        with suppress(TypeError):
            return integral_fn(x=x, limits=limits, norm_range=norm, params=params, model=model)
        with suppress(TypeError):
            return integral_fn(limits=limits, norm_range=norm, params=params, model=model)
        msg = "Could not integrate, unknown reason. Please fill a bug report."
        raise AssertionError(msg)

        # with suppress(TypeError):
        #     return integral_fn(x=x, limits=limits, norm=norm, params=params, model=model)
        # with suppress(TypeError):
        #     return integral_fn(limits=limits, norm=norm, params=params, model=model)


class Integral:  # TODO analytic integral
    def __init__(self, func: Callable, limits: ZfitSpace, priority: int | float):
        """A lightweight holder for the integral function."""
        self.limits = limits
        self.integrate = func
        self.axes = limits.axes
        self.priority = priority

    def __call__(self, *args, **kwargs):
        return self.integrate(*args, **kwargs)


# to be "the" future integral class
class Integration:
    def __init__(self, mc_sampler, draws_per_dim, tol, max_draws, draws_simpson):
        self.tol = tol
        self.max_draws = max_draws
        self.mc_sampler = mc_sampler
        self.draws_per_dim = draws_per_dim
        self.draws_simpson = draws_simpson



#  Copyright (c) 2024 zfit
import numpy as np
import zfit
from zfit import z
import zfit.z.numpy as znp
from scipy.special import voigt_profile


def test_voigt1():
    mean1_true = 1.0
    std1_true = 2.0
    width1_true = 1.4

    obs1 = "obs1"
    limits1 = (-10, 10)
    obs1 = zfit.Space(obs1, limits=limits1)

    test_values = z.random.uniform(minval=-10, maxval=10, shape=(1000,))

    mean = zfit.Parameter("mean", mean1_true, -10, 10)
    sigma = zfit.Parameter("sigma", std1_true, 0.1, 10)
    gamma = zfit.Parameter("gamma", width1_true, 0.1, 10)
    voigt1 = zfit.pdf.Voigt(m=mean, sigma=sigma, gamma=gamma, obs=obs1)

    probs1 = voigt1.pdf(x=test_values, norm=False)
    np.testing.assert_array_less(0, probs1)
    sample1 = voigt1.sample(100)
    assert len(sample1.value()) == 100

    probs2 = voigt_profile(test_values - mean1_true, std1_true, width1_true)
    np.testing.assert_allclose(probs1, probs2, rtol=1e-10, atol=1e-08)

    func = lambda: znp.sum(
        zfit.pdf.Voigt(m=mean, sigma=sigma, gamma=gamma, obs=obs1).pdf(x=test_values)
    )

    num_gradients = z.math.numerical_gradient(func, params=[mean, sigma, gamma])
    tf_gradients = z.math.autodiff_gradient(func, params=[mean, sigma, gamma])
    np.testing.assert_allclose(num_gradients, tf_gradients, rtol=1e-6)



#  Copyright (c) 2024 zfit

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

from collections.abc import Iterable

import tensorflow as tf

from ..core.interfaces import ZfitIndependentParameter, ZfitLoss
from ..util.exception import OperationNotAllowedError
from .baseminimizer import BaseStepMinimizer, minimize_supports


class WrapOptimizer(BaseStepMinimizer):
    def __init__(
        self,
        optimizer,
        tol=None,
        criterion=None,
        strategy=None,
        verbosity=None,
        name=None,
        **kwargs,
    ):
        """Wrap TensorFlow optimizers to have the zfit interface and behavior.

        .. warning:: The optimizers changed the interface in Keras 3, which is not (yet) supported.

          As they tend to perform significantly worse than the optimizers in `scipy` or `iminuit`, it is
            not a priority to support them, if ever at all. If you need them, please open an issue/PR: https://github.com/zfit/zfit.

        .. note:: Different behavior of minimize

          While the ``minimize`` method in TensorFlow optimizers executes a single step of the minimization,
          the ``minimize`` method of a :class:`ZfitMinimizer` fully minimizes a function until convergence
          is reached. In order to execute a single step, use the :meth:`~WrapOptimizer.step` method (however,
          this is in general not necessary to do and rather inefficient).

        Args:
            optimizer: Instance of a :class:`tf.keras.optimizers.Optimizer`.
            All other arguments: these are passed through and have the same functionality as described
            in :py:class:`~BaseStepMinimizer`
        """
        if not isinstance(optimizer, tf.keras.optimizers.Optimizer):
            msg = f"optimizer {optimizer!s} has to be from class Optimizer"
            raise TypeError(msg)
        if not hasattr(optimizer, "minimize"):
            msg = (
                f"optimizer {optimizer!s} has to have a method minimize and is most probably the new Keras 3 interface that is not yet supported."
                "Please open an issue/PR if you would need TF/Keras optimizers, as they generally don't perform well anyway."
            )
            raise AttributeError(msg)
        super().__init__(
            tol=tol,
            criterion=criterion,
            strategy=strategy,
            verbosity=verbosity,
            name=name,
            minimizer_options=None,
            **kwargs,
        )
        self._optimizer_tf = optimizer

    @minimize_supports(init=True)
    def _minimize(self, loss, params, init):
        try:
            return super()._minimize(loss, params, init)
        except ValueError as error:
            if "No gradients provided for any variable" in error.args[0]:
                msg = "Cannot use TF optimizer with" " a numerical gradient (non-TF function)"
                raise OperationNotAllowedError(msg) from None
            raise

    def _step(
        self,
        loss: ZfitLoss,
        params: Iterable[ZfitIndependentParameter],
        init: zfit.result.FitResult | None,
    ) -> tf.Tensor:
        del init  # unused
        # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
        # import zfit
        # params = zfit.z.math._extract_tfparams(params)
        value = partial(loss.value, full=False)
        self._optimizer_tf.minimize(loss=value, var_list=params)
        return value()



#  Copyright (c) 2024 zfit
import hist.axis
import matplotlib.pyplot as plt
import mplhep
import numpy as np
import pytest
import scipy.stats

import zfit
import zfit.z.numpy as znp
from zfit.util.exception import SpecificFunctionNotImplemented


def create_gaussians(yields):
    obs = zfit.Space("obs1", (-2, 5))

    gauss1 = zfit.pdf.Gauss(mu=2, sigma=4, obs=obs)
    gauss2 = zfit.pdf.Gauss(mu=3, sigma=5, obs=obs)
    gauss3 = zfit.pdf.Gauss(mu=1, sigma=6, obs=obs)
    gausses = gauss1, gauss2, gauss3

    if yields:
        if len(yields) == 2:
            yields = yields + [None]
        gausses = [
            gauss.create_extended(yiel) if yiel is not None else gauss
            for gauss, yiel in zip(gausses, yields)
        ]
    return gausses


def test_analytic_integral():
    obs = zfit.Space("obs1", (-2, 5))


@pytest.mark.parametrize("yields", [False, [4, 3, 5], [1, 2]])
def test_frac_behavior(yields):
    gauss1, gauss2, gauss3 = create_gaussians(yields)

    frac1 = zfit.Parameter("frac1", 0.4)
    frac2 = zfit.Parameter("frac2", 0.6)
    for fracs in [frac1, [frac1, frac2]]:
        sumpdf1 = zfit.pdf.SumPDF([gauss1, gauss2], fracs)
        assert sumpdf1.fracs[0] == frac1
        assert len(sumpdf1.fracs) == 2
        assert len(sumpdf1.params) == 2
        assert sumpdf1.params["frac_0"] == frac1
        assert sumpdf1.params["frac_1"] == sumpdf1.fracs[1]
        assert not sumpdf1.is_extended

        frac2_val = 1 - frac1.value()
        assert (
            pytest.approx(frac2_val) == sumpdf1.params["frac_1"].value()
        )
        if isinstance(fracs, list) and len(fracs) == 2:
            assert sumpdf1.params["frac_1"] == frac2

    frac1.set_value(0.3)
    frac3 = zfit.Parameter("frac3", 0.1)

    for fracs in [[frac1, frac2], [frac1, frac2, frac3]]:
        sumpdf2 = zfit.pdf.SumPDF([gauss1, gauss2, gauss3], fracs)
        assert sumpdf2.fracs[0] == frac1
        assert len(sumpdf2.fracs) == 3
        assert len(sumpdf2.params) == 3
        assert sumpdf2.params["frac_0"] == frac1
        assert sumpdf2.params["frac_1"] == frac2
        assert (
            pytest.approx(frac3.value())
            == sumpdf2.params["frac_2"].value()
        )
        assert not sumpdf1.is_extended

        if isinstance(fracs, list) and len(fracs) == 3:
            assert sumpdf2.params["frac_2"] == frac3


@pytest.mark.flaky(2)  # ks test
def test_sampling():
    class SimpleSampleSumPDF(zfit.pdf.SumPDF):
        @zfit.supports()
        def _sample(self, n, limits):
            raise SpecificFunctionNotImplemented  # fallback to the default sampling

    sample_size = 100_000
    tol = 0.1
    mu1, mu2 = 0, 10
    frac = 0.9
    true_mu = mu1 * frac + mu2 * (1 - frac)

    obs = zfit.Space("obs1", (mu1 - 5, mu2 + 5))
    gauss1 = zfit.pdf.Gauss(obs=obs, mu=mu1, sigma=1)
    gauss2 = zfit.pdf.Gauss(obs=obs, mu=mu2, sigma=1)

    sumpdf = zfit.pdf.SumPDF([gauss1, gauss2], frac)
    sumpdf_true = SimpleSampleSumPDF([gauss1, gauss2], frac)

    sample = sumpdf.sample(sample_size).value()[:, 0]
    sample_true = sumpdf_true.sample(sample_size).value()[:, 0]

    assert pytest.approx(
        np.mean(sample_true), abs=tol
    ) == true_mu  # if this is not True, it's a problem, the test is flawed
    assert pytest.approx(np.mean(sample), abs=tol) == true_mu
    assert pytest.approx(np.std(sample), abs=tol) == np.std(sample_true)

    plt.figure()
    plt.title("Sampling of SumPDF")
    axis = hist.axis.Regular(100, *obs.v1.limits)
    sample_hist = hist.Hist(axis).fill(sample)
    sample_true_hist = hist.Hist(axis).fill(sample_true)
    mplhep.histplot(sample_hist, density=True, label="sampled")
    mplhep.histplot(sample_true_hist, density=True, label="sampled true")
    plt.legend()
    pytest.zfit_savefig(folder="sampling")
    assert scipy.stats.mannwhitneyu(sample, sample_true).pvalue > 0.05


@pytest.mark.flaky(2)  # mc integration
def test_integrate():
    class SimpleSampleSumPDF(zfit.pdf.SumPDF):
        @zfit.supports()
        def _integrate(self, limits, norm, options):
            raise SpecificFunctionNotImplemented  # force fallback to the default sampling

        @zfit.supports()
        def _analytic_integrate(self, limits, norm):
            raise SpecificFunctionNotImplemented

        @zfit.supports()
        def _numeric_integrate(self, limits, norm=None, options=None):
            raise SpecificFunctionNotImplemented

    mu1, mu2 = 0, 1.7
    frac = 0.7

    lower = mu1 - 2.5
    upper = mu2 + 3
    obs = zfit.Space("obs1", (lower, upper))
    limits = zfit.Space("obs1", (mu1 - 0.3, mu2 + 0.1))
    gauss1 = zfit.pdf.Gauss(obs=obs, mu=mu1, sigma=0.93)
    gauss2 = zfit.pdf.Gauss(obs=obs, mu=mu2, sigma=1.2)

    sumpdf = zfit.pdf.SumPDF([gauss1, gauss2], frac)
    sumpdf_true = SimpleSampleSumPDF([gauss1, gauss2], frac)
    assert pytest.approx(1, abs=0.01) == (
        gauss1.integrate(
            limits,
            norm=limits,
        )
    )
    integral = sumpdf.integrate(
        limits=limits,
        norm=False,
    )
    integral_true = sumpdf_true.integrate(
        limits=limits,
        norm=False,
    )
    integral_manual_true = gauss1.integrate(
        limits,
    ) * frac + gauss2.integrate(
        limits,
    ) * (1 - frac)

    assert pytest.approx(integral_manual_true, rel=0.03) == integral_true
    assert pytest.approx(integral, rel=0.03) == integral_true
    assert integral_true < 0.85

    analytic_integral = sumpdf.analytic_integrate(limits=limits, norm=False)

    assert pytest.approx(analytic_integral, rel=0.03) == integral_true

    rnd_limits = [lower] + sorted(np.random.uniform(lower, upper, 16)) + [upper]
    integrals = [
        sumpdf.integrate((low, up), norm=False)
        for low, up in zip(rnd_limits[:-1], rnd_limits[1:])
    ]

    integral = np.sum(integrals)
    integral_full = sumpdf.integrate((lower, upper), norm=False)
    assert pytest.approx(integral_full) == integral



#  Copyright (c) 2024 zfit
from __future__ import annotations

from ._data.binneddatav1 import BinnedData, BinnedSamplerData
from ._variables.axis import RegularBinning, VariableBinning
from .core.data import Data, SamplerData, concat, convert_to_data

__all__ = [
    "Data",
    "BinnedData",
    "RegularBinning",
    "VariableBinning",
    "convert_to_data",
    "SamplerData",
    "BinnedSamplerData",
    "concat",
]



#  Copyright (c) 2024 zfit
from __future__ import annotations

import matplotlib.pyplot as plt
import mplhep
import numpy as np

import zfit

plt.style.use(mplhep.style.LHCb2)
n_bins = 50
# create space
obs_binned = zfit.Space("x", binning=zfit.binned.RegularBinning(50, -10, 10, name="x"))
obs = obs_binned.with_binning(None)  # unbinned obs

# parameters, model building, pdf creation
mu = zfit.Parameter("mu", 1.0, -4, 6)
sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)
nr = zfit.Parameter("nr", 2, 0, 5)
alpha = zfit.Parameter("alpha", 1.5, 0, 5)
n_sig = zfit.Parameter("n_sig", 5000 * 0.3)
doublecb = zfit.pdf.GeneralizedCB(
    mu=mu,
    sigmal=sigma,
    nl=nr,
    alphal=alpha,  # sharing sigma, meaning the width is the same left and right
    sigmar=sigma,
    nr=nr,
    alphar=0.5,
    obs=obs,
    extended=n_sig,
)

lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
n_bkg = zfit.Parameter("n_bkg", 5000 * (1 - 0.3))
exponential = zfit.pdf.Exponential(lambd, obs=obs, extended=n_bkg)

# since `obs_binned` is binned, giving it to the pdf will return a binned pdf
# note that this is different from binning both pdfs and summing them (there, we would need a BinnedSumPDF)
model = zfit.pdf.SumPDF([doublecb, exponential], obs=obs_binned)

# data
n_sample = 5000
data = model.sample(n=n_sample)

plot_scaling = n_sample / n_bins * obs.volume

x = np.linspace(obs.v1.lower, obs.v1.upper, 1000)


def plot_pdf(title):
    plt.figure()
    plt.title(title)
    y = model.pdf(x)
    y_doublecb = doublecb.pdf(x) * model.pdfs[0].params["frac_0"]
    y_exp = exponential.pdf(x) * model.pdfs[0].params["frac_1"]
    plt.plot(x, y * plot_scaling, label="Sum - Binned Model")
    plt.plot(x, y_doublecb * plot_scaling, label="Gauss - Signal")
    plt.plot(x, y_exp * plot_scaling, label="Exp - Background")
    # mplhep.histplot(np.histogram(data_np, bins=n_bins), yerr=True, color='black', histtype='errorbar')
    mplhep.histplot(data, yerr=data.counts() ** 0.5, color="black", histtype="errorbar")
    plt.ylabel("Counts")
    plt.xlabel("obs: $B_{mass}$")
    plt.legend()
    # plt.savefig(title + ".pdf")  # uncomment to save the plot


# set the values to a start value for the fit
zfit.param.set_values({mu: 0.5, sigma: 1.2, nr: 2.0, alpha: 1.5, lambd: -0.05})
# alternatively, we can set the values with a list/array
# zfit.param.set_values([mu, sigma, nr, alpha, lambd], [0.5, 1.2, 2.0, 1.5, -0.05])

# create NLL
nll = zfit.loss.ExtendedBinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit(gradient="zfit")

plot_pdf("before fit")

result = minimizer.minimize(nll).update_params()
# do the error calculations, here with hesse, than with minos

param_hesse = result.hesse()
(
    param_errors,
    _,
) = result.errors()  # this returns a new FitResult if a new minimum was found

# this returns a new FitResult if a new minimum was found
param_errors, _ = result.errors()
# plot the data and pdf
plot_pdf("after fit")
# uncomment to display plots
# plt.show()



#  Copyright (c) 2024 zfit
import numpy as np


def test_pdf_lognormal():
    import zfit

    mu = zfit.Parameter("mu", 3.0)
    sigma = zfit.Parameter("sigma", 1.0)
    obs = zfit.Space("obs1", limits=(0, 10))
    lognormal = zfit.pdf.LogNormal(mu=mu, sigma=sigma, obs=obs)
    probs1 = lognormal.pdf(x=[0.1, 3, 6, 9])
    assert np.all(probs1 >= 0)
    assert np.all(np.isfinite(probs1))
    sample1 = lognormal.sample(1000)
    assert len(sample1.value()) == 1000
    probs2 = lognormal.pdf(x=sample1)
    assert np.all(probs2 >= 0)



#  Copyright (c) 2024 zfit

import numpy as np
import pytest

import zfit
import zfit.z.numpy as znp
import tensorflow as tf

from zfit import supports, z
from zfit.models.cache import CachedPDF
from zfit.util import ztyping
from zfit.util.exception import AnalyticGradientNotAvailable


class TestPDF(zfit.pdf.BaseFunctor):
    def __init__(self, obs: ztyping.ObsTypeInput, mu, sigma):
        gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
        super().__init__(pdfs=gauss, obs=obs)
        self.pdf_call_counter = tf.Variable(0)
        self.integrate_call_counter = tf.Variable(0)

    @supports(norm="space")
    def _pdf(self, x, norm, *, norm_range=None):
        self.pdf_call_counter.assign_add(1)
        return self.pdfs[0].pdf(x)

    @supports(norm="space")
    def _integrate(self, limits, norm, options=None):
        self.integrate_call_counter.assign_add(1)
        return self.pdfs[0].integrate(limits)


def test_cached_pdf_equals_pdf_without_cache():

    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    cached_gauss = CachedPDF(gauss)
    x = znp.linspace(-5, 5, 500)
    xhashed = zfit.data.Data.from_numpy(obs=obs, array=x, use_hash=True)
    assert xhashed.hashint is not None
    xunhashed = zfit.data.Data.from_numpy(obs=obs, array=x, use_hash=False)
    assert xunhashed.hashint is None
    assert tf.math.reduce_all(tf.equal(gauss.pdf(xhashed), cached_gauss.pdf(xhashed)))
    # hash not yet used inside cached pdf
    # with pytest.raises(ValueError):
    #     _ = gauss.pdf(xunhashed)


def test_pdf_cache_is_used():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    x = znp.linspace(-5, 5, 500)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(0))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(1))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(1))


def test_pdf_cache_revaluation_if_mu_was_changed():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    x = znp.linspace(-5, 5, 500)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(0))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(1))
    mu.set_value(2.0)
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))


def test_pdf_cache_revaluation_if_sigma_was_changed():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    x = znp.linspace(-5, 5, 500)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(0))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(1))
    sigma.set_value(3.0)
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))


def test_pdf_cache_revaluation_if_x_was_changed():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    x = znp.linspace(-5, 5, 500)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(0))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(1))
    x = znp.linspace(-5, 0, 500)
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))
    cached_test_pdf.pdf(x)
    assert tf.equal(test_pdf.pdf_call_counter, tf.Variable(2))


def test_cached_integrate_equals_integrate_without_cache():
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    cached_gauss = CachedPDF(gauss)

    assert tf.math.reduce_all(
        tf.equal(gauss.integrate(limits=obs), cached_gauss.integrate(limits=obs))
    )


def test_integrate_cache_is_used():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(0))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(1))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(1))


def test_integrate_cache_is_revaluation_if_mu_was_changed():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(0))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(1))
    mu.set_value(2.0)
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))


def test_integrate_cache_is_revaluation_if_sigma_was_changed():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(0))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(1))
    sigma.set_value(3.0)
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))


def test_integrate_cache_is_revaluation_if_limits_is_different():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(0))
    cached_test_pdf.integrate(limits=obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(1))
    obs = zfit.Space("x", limits=[-5.0, 0.0])
    cached_test_pdf.integrate(obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))
    cached_test_pdf.integrate(obs)
    assert tf.equal(test_pdf.integrate_call_counter, tf.Variable(2))


def test_gradient_cached_pdf():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf)
    x = znp.linspace(-5, 5, 500)
    with tf.GradientTape(watch_accessed_variables=True, persistent=True) as tape:

        pdf = cached_test_pdf.pdf(x)

    with pytest.raises(AnalyticGradientNotAvailable):
        _ = tape.gradient(pdf, [mu])


def test_minimize_cached_pdf():
    obs = zfit.Space("x", limits=[-5.0, 5.0])
    mu = zfit.Parameter("mu", 1.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1, 0, 10)
    test_pdf1 = TestPDF(obs=obs, mu=mu, sigma=sigma)
    cached_test_pdf = CachedPDF(test_pdf1)

    mu2 = zfit.Parameter("mu2", 5.0, -5, 5)
    sigma2 = zfit.Parameter("sigma2", 3, 0, 10)
    test_pdf2 = TestPDF(obs=obs, mu=mu2, sigma=sigma2)
    cached_test_pdf2 = CachedPDF(test_pdf2)

    testpdf = zfit.pdf.SumPDF([cached_test_pdf, cached_test_pdf2], fracs=0.5)

    array1 = np.random.normal(1.3, 0.8, 5000)
    array2 = np.random.normal(3.4, 1.3, 5000)
    array = znp.concatenate([array1, array2])
    data = zfit.Data.from_numpy(obs=obs, array=array)
    nll = zfit.loss.UnbinnedNLL(model=testpdf, data=data)
    minimizer = zfit.minimize.Minuit(gradient="zfit")
    with pytest.raises(AnalyticGradientNotAvailable):
        _ = minimizer.minimize(nll)
    minimizer = zfit.minimize.Minuit(gradient=True)
    result = minimizer.minimize(nll)
    assert result.converged
    assert result.valid
    result.hesse()



#  Copyright (c) 2024 zfit
from __future__ import annotations

from typing import Iterable, Literal, Union

import numpy as np
import pydantic
import tensorflow as tf

import zfit.z.numpy as znp

from .. import z
from ..core.interfaces import ZfitPDF, ZfitSpace
from ..core.serialmixin import SerializableMixin
from ..core.space import Space, convert_to_space, supports
from ..serialization import Serializer, SpaceRepr  # noqa: F401
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import SpecificFunctionNotImplemented
from .basefunctor import FunctorPDFRepr
from .functor import BaseFunctor


def check_limits(limits: Union[ZfitSpace, list[ZfitSpace]], obs=None):
    """Check if the limits are valid Spaces and return an iterable."""
    limits = convert_to_container(limits, container=tuple)
    obs = obs.obs if obs is not None else None

    newlimits = []
    if limits is not None:
        # check if it's a list of exactly two limits given as numbers or arrays
        if (
            obs is not None
            and (
                np.shape(limits) == (2, len(obs))
                or (
                    len(limits) == 2
                    and np.atleast_1d(limits[0]).shape == (len(obs),)
                    and np.atleast_1d(limits[1]).shape == (len(obs),)
                )
            )
            and not (isinstance(limits[0], ZfitSpace) or isinstance(limits[1], ZfitSpace))
        ):
            limits = [Space(obs=obs, limits=limits)]
        notspace = []
        for limit in limits:
            if not isinstance(limit, ZfitSpace):
                if obs is None:
                    notspace.append(limit)
                else:
                    limit = Space(obs=obs, limits=limit)
            newlimits.append(limit)
        limits = newlimits
        if notspace:
            msg = f"limits {notspace} are not of type ZfitSpace and obs could not be automatically determined."
            raise TypeError(msg)
    limits_sorted = tuple(sorted(limits, key=lambda limit: limit.v1.lower))
    return check_overlap(limits_sorted)


def check_overlap(limits):
    if len(limits) == 1:
        return limits
    for i, limit1 in enumerate(limits[:-1]):
        for limit2 in limits[i + 1 :]:
            if limit1.v1.upper > limit2.v1.lower:
                msg = f"Limit {limit1} overlaps with {limit2} in TruncatedPDF."
                raise ValueError(msg)
    return limits


# TODO: implement smart limits
# def create_subset_limits(limits, constraints):
#     limits = check_limits(limits)
#     newlimits = []
#     obs = limits[0].obs
#     axes = obs.axis
#     for limit in limits:
#         if not limit.ndims == 1:
#             raise ValueError(f"Limit {limit} is not 1-dimensional.")
#         newlower, newupper = None, None
#         for constr in constraints:
#             if constr.v1.lower <= limit.v1.lower <= constr.v1.upper:
#                 assert newlower is None, "Multiple limits overlap with the same limit, should have been caught before. All limits: {limits}"
#                 newlower = max(limit.v1.lower, constr.v1.lower)
#             if newlower and constr.v1.lower <= limit.v1.upper <= constr.v1.upper:
#                 assert newupper is None, "Multiple limits overlap with the same limit, should have been caught before. All limits: {limits}"
#                 newupper = min(limit.v1.upper, constr.v1.upper)
#         if newlower is not None and newupper is not None:
#             newlimits.append(Space(obs=obs, lower=newlower, upper=newupper, axes=axes))
#             break


class TruncatedPDF(BaseFunctor, SerializableMixin):
    def __init__(
        self,
        pdf: ZfitPDF,
        limits: ZfitSpace | Iterable[ZfitSpace],
        obs: ztyping.ObsTypeInput = None,
        *,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormRangeTypeInput = None,
        name: str | None = None,
        label: str | None = None,
    ):
        """Truncated PDF in one or multiple ranges.

        The PDF is truncated to the given limits, i.e. the PDF is only evaluated within the given limits
        and outside the limits, the PDF is zero. The limits can be given as a single or as a list of limits.
        This solves two problems: first, the PDF is only evaluated where it is defined and zero outside,
        which enables PDFs that are not defined outside the limits to be used (Poisson, LogNormal, Chi2,...).
        Second, the list of limits can be used to effectively create a multilimit PDF, that means, a PDF that
        is defined in multiple ranges and zero outside these ranges, also in between.

        Historically, this was implemented in zfit through MultipleLimits, which is now deprecated and replaced
        by this class.

        Args:
            pdf: The PDF to be truncated.
            limits: The limits to truncate the PDF. Can be a single limit or multiple limits.
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
               If None, the PDF will be extended if the original PDF is extended.
               If ``True`` and the original PDF is extended, the yield will be scaled to the
               fraction of the total integral that is within the limits.
               Therefore, the overall yield is comparable, i.e. the pdfs can be plotted
               "on top of each other".
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        original_init = {"extended": extended, "obs": obs}
        if name is None:
            name = "TruncatedPDF"
        obs = pdf.space if obs is None else convert_to_space(obs)
        self._limits = check_limits(limits, obs=obs)
        if extended is None:
            extended = pdf.is_extended
        if extended is True and pdf.is_extended:
            base_norm = pdf.integrate(limits=obs, norm=False)
            piecewise_norms = znp.asarray([pdf.integrate(limits=limit, norm=False) for limit in self._limits])
            relative_scale = znp.sum(piecewise_norms / base_norm)
            import zfit

            extended = zfit.param.ComposedParameter(
                name=f"AUTO_yield{zfit.core.parameter.get_auto_number()!s}_{name}",
                func=lambda params, scale=relative_scale: params["wrapped_yield"] * scale,
                params={"wrapped_yield": pdf.get_yield()},
            )
        super().__init__(obs=obs, name=name, extended=extended, norm=norm, pdfs=pdf, label=label)
        if self.obs != pdf.obs:
            msg = f"The space of the TruncatedPDF ({self.obs}) must be the same as the PDF ({pdf.space})."
            raise ValueError(msg)
        self.hs3.original_init.update(original_init)

    @property
    def limits(self):
        return self._limits

    def _unnormalized_pdf(self, x):
        # the implementation only feeds the pdf with data that is inside the limits that we want to evaluate
        # this maybe has a speedup (although limited as we're using dynamic shapes -> needs to change for JAX)
        # but most importantly avoids any issue if we take the gradient of a pdf that is not defined outside
        # the limits, because this can yield NaNs using a naive multiplication with a mask
        from zfit import Data

        xarray = znp.asarray(x.value())
        inside_arrays = [limit.inside(x) for limit in self._limits]
        inside = znp.any(inside_arrays, axis=0)
        indices = znp.transpose(znp.where(inside))  # is a list? but transpose gives perfect shape for indices
        data = Data.from_tensor(tensor=xarray[inside], obs=self.obs)
        prob = self.pdfs[0].pdf(data, norm=False)
        return tf.scatter_nd(indices, prob, tf.shape(xarray, out_type=np.int64)[:1])  # only nevents

    # todo: not needed? We just allow truncation to have multiple limits, should be sufficient?
    # @supports(norm=True)
    # def _normalization(self, norm, options):
    #     if (norms := self._norms) is None:
    #         raise SpecificFunctionNotImplemented("Fallback to default, no norms given.")
    #     elif norm != self.space:
    #         msg = f"Cannot normalize to a different space than the one given, the norms {norms}."
    #         raise SpecificFunctionNotImplemented(msg)
    #
    #
    #     normterms = [self.normalization(norm, options=options) for norm in norms]
    #     return znp.sum(normterms, axis=0)

    @supports()
    def _integrate(self, limits, norm, options=None):
        del norm  # not used here
        if limits != self.space:  # we could also do it, but would need to check each limit
            raise SpecificFunctionNotImplemented
        limits = convert_to_container(
            self.limits
        )  # if it's the overarching limits, we can just use our own ones, the real ones
        # limits = create_subset_limits(limits, self.limits)  # TODO: be smart about limits, we would not need to throw the SpecificFunctionNotImplemented
        integrals = [self.pdfs[0].integrate(limits=limit, norm=False, options=options) for limit in limits]
        return znp.sum(integrals, axis=0)

    # TODO: we could make sampling more efficient by only sampling the relevant ranges, however, that would
    # mean we need to check if the limits of the pdf are within the limits given
    @supports()
    def _sample(self, n, limits):
        pdf = self.pdfs[0]
        if limits != self.space:  # we could also do it, but would need to check each limit
            raise SpecificFunctionNotImplemented
        limits = self.limits
        # should be `self.integrate`, but as we do it numerically currently, more efficient to use pdf
        if len(limits) > 1:
            integrals = znp.concatenate([pdf.integrate(limits=limit, norm=False) for limit in limits])
            fracs = integrals / znp.sum(integrals, axis=0)  # norm
            fracs.set_shape([len(limits)])
            counts = tf.unstack(z.random.counts_multinomial(n, probs=fracs), axis=0)
        else:
            counts = [n]
        samples = [self.pdfs[0].sample(count, limits=limit).value() for count, limit in zip(counts, limits)]
        return znp.concatenate(samples, axis=0)


class TruncatedPDFRepr(FunctorPDFRepr):
    _implementation = TruncatedPDF
    hs3_type: Literal["TruncatedPDF"] = pydantic.Field("TruncatedPDF", alias="type")
    limits: list[SpaceRepr]

    def _to_orm(self, init):
        init["pdf"] = init.pop("pdfs")[0]
        return super()._to_orm(init)
